from __future__ import annotations

import warnings
from abc import ABC, abstractmethod
from contextlib import contextmanager
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Iterable,
    Iterator,
    Literal,
    Sequence,
    TypeVar,
    overload,
)

import numpy as np
from cmap import Color
from numpy.typing import ArrayLike
from psygnal import Signal, SignalGroup

from whitecanvas import layers as _l
from whitecanvas import protocols, theme
from whitecanvas.backend import Backend, patch_dummy_backend
from whitecanvas.canvas import _namespaces as _ns
from whitecanvas.canvas import dataframe as _df
from whitecanvas.canvas import layerlist as _ll
from whitecanvas.canvas._between import BetweenPlotter
from whitecanvas.canvas._dims import Dims
from whitecanvas.canvas._fit import FitPlotter
from whitecanvas.canvas._palette import ColorPalette
from whitecanvas.canvas._stacked import StackOverPlotter
from whitecanvas.layers import _legend, _mixin
from whitecanvas.layers import group as _lg
from whitecanvas.types import (
    Alignment,
    ArrayLike1D,
    ColormapType,
    ColorType,
    Hatch,
    HistBinType,
    KdeBandWidthType,
    LineStyle,
    Location,
    LocationStr,
    Orientation,
    OrientationLike,
    Rect,
    StepStyle,
    StepStyleStr,
    Symbol,
)
from whitecanvas.utils.normalize import as_array_1d, normalize_xy
from whitecanvas.utils.type_check import is_real_number

if TYPE_CHECKING:
    from typing_extensions import Concatenate, ParamSpec, Self

    _P = ParamSpec("_P")
    _DF = TypeVar("_DF")

_L = TypeVar("_L", bound=_l.Layer)
_L0 = TypeVar("_L0", _l.Bars, _l.Band)


class CanvasEvents(SignalGroup):
    lims = Signal(Rect)
    drawn = Signal()


class CanvasNDBase(ABC):
    title = _ns.TitleNamespace()
    x = _ns.XAxisNamespace()
    y = _ns.YAxisNamespace()
    layers = _ll.LayerList()
    events: CanvasEvents

    def __init__(self, palette: ColormapType | None = None):
        if palette is None:
            palette = theme.get_theme().palette
        self.events = CanvasEvents()
        self._color_palette = ColorPalette(palette)
        self._is_grouping = False

    @abstractmethod
    def _get_backend(self) -> Backend:
        """Return the backend."""

    @abstractmethod
    def _canvas(self) -> protocols.CanvasProtocol:
        """Return the canvas object."""

    @abstractmethod
    def _autoscale_for_layer(self, layer: _l.Layer):
        """Autoscale the canvas for the given layer."""

    def _draw_canvas(self):
        self._canvas()._plt_draw()
        self.events.drawn.emit()

    def _coerce_name(self, name: str | None, default: str = "data") -> str:
        if name is None:
            basename = default
            name = f"{default}-0"
        else:
            basename = name
        i = 0
        _exists = {layer.name for layer in self.layers}
        while name in _exists:
            name = f"{basename}-{i}"
            i += 1
        return name

    def _cb_inserted(self, idx: int, layer: _l.Layer):
        if self._is_grouping:
            # this happens when the grouped layer is inserted
            layer._connect_canvas(self)
            return

        _canvas = self._canvas()
        for l in _iter_layers(layer):
            _canvas._plt_add_layer(l._backend)
            l._connect_canvas(self)

        if isinstance(layer, _l.LayerWrapper):
            # TODO: check if connecting LayerGroup is necessary
            layer._connect_canvas(self)
        # autoscale
        self._autoscale_for_layer(layer)
        self._cb_reordered()

    def _cb_reordered(self):
        layer_backends = []
        for layer in self.layers:
            if isinstance(layer, _l.PrimitiveLayer):
                layer_backends.append(layer._backend)
            elif isinstance(layer, _l.LayerGroup):
                for child in layer.iter_primitive():
                    layer_backends.append(child._backend)
            elif isinstance(layer, _l.LayerWrapper):
                for child in _iter_layers(layer):
                    layer_backends.append(child._backend)
            else:
                raise RuntimeError(f"type {type(layer)} not expected")
        self._canvas()._plt_reorder_layers(layer_backends)

    def _cb_removed(self, idx: int, layer: _l.Layer):
        if self._is_grouping:
            return
        _canvas = self._canvas()
        for l in _iter_layers(layer):
            _canvas._plt_remove_layer(l._backend)
            l._disconnect_canvas(self)

    def _cb_layer_grouped(self, group: _l.LayerGroup):
        indices: list[int] = []  # layers to remove
        not_found: list[_l.PrimitiveLayer] = []  # primitive layers to add
        id_exists = set(map(id, self.layers.iter_primitives()))
        for layer in group.iter_children():
            try:
                idx = self.layers.index(layer)
                indices.append(idx)
            except ValueError:
                not_found.extend(_iter_layers(layer))
        if not indices:
            return
        self._is_grouping = True
        try:
            for idx in reversed(indices):
                # remove from the layer list since it is directly grouped
                self.layers.pop(idx)
            self.layers.append(group)
            _canvas = self._canvas()
            for child in not_found:
                if id(child) in id_exists:
                    # skip since it is already in the canvas
                    continue
                child._connect_canvas(self)
                _canvas._plt_add_layer(child._backend)
        finally:
            self._is_grouping = False
        self._cb_reordered()
        self._autoscale_for_layer(group)

    def _generate_colors(self, color: ColorType | None) -> Color:
        if color is None:
            color = self._color_palette.next()
        return color

    @property
    def autoscale_enabled(self) -> bool:
        """Return whether autoscale is enabled."""
        return self._autoscale_enabled

    @autoscale_enabled.setter
    def autoscale_enabled(self, enabled: bool):
        if not isinstance(enabled, bool):
            raise TypeError(f"Expected a bool, got {type(enabled)}.")
        self._autoscale_enabled = enabled

    @contextmanager
    def autoscale_context(self, enabled: bool):
        """Context manager to temporarily change the autoscale state."""
        _was_enabled = self.autoscale_enabled
        self.autoscale_enabled = enabled
        try:
            yield
        finally:
            self.autoscale_enabled = _was_enabled


class CanvasBase(CanvasNDBase):
    """Base class for any canvas object."""

    dims = Dims()
    overlays = _ll.LayerList()
    mouse = _ns.MouseNamespace()

    def __init__(self, palette: ColormapType | None = None):
        super().__init__(palette)
        self._autoscale_enabled = True
        if not self._get_backend().is_dummy():
            self._init_canvas()

    def _init_canvas(self):
        # default colors and font
        _t = theme.get_theme()
        _ft = _t.font
        self.x.color = _t.foreground_color
        self.y.color = _t.foreground_color
        self.x.label.update(family=_ft.family, color=_ft.color, size=_ft.size)
        self.y.label.update(family=_ft.family, color=_ft.color, size=_ft.size)
        self.title.update(family=_ft.family, color=_ft.color, size=_ft.size)
        self.x.ticks.update(family=_ft.family, color=_ft.color, size=_ft.size)
        self.y.ticks.update(family=_ft.family, color=_ft.color, size=_ft.size)

        # connect layer events
        self.layers.events.inserted.connect(
            self._cb_inserted, unique=True, max_args=None
        )
        self.layers.events.removed.connect(self._cb_removed, unique=True, max_args=None)
        self.layers.events.reordered.connect(
            self._cb_reordered, unique=True, max_args=None
        )
        self.layers.events.connect(self._draw_canvas, unique=True, max_args=None)

        self.overlays.events.inserted.connect(
            self._cb_overlay_inserted, unique=True, max_args=None
        )
        self.overlays.events.removed.connect(
            self._cb_removed, unique=True, max_args=None
        )
        self.overlays.events.connect(self._draw_canvas, unique=True, max_args=None)

        canvas = self._canvas()
        canvas._plt_connect_xlim_changed(self._emit_xlim_changed)
        canvas._plt_connect_ylim_changed(self._emit_ylim_changed)

        if hasattr(canvas, "_plt_canvas_hook"):
            canvas._plt_canvas_hook(self)

    def _install_mouse_events(self):
        canvas = self._canvas()
        canvas._plt_connect_mouse_click(self.mouse.clicked.emit)
        canvas._plt_connect_mouse_click(self.mouse.moved.emit)
        canvas._plt_connect_mouse_drag(self.mouse.moved.emit)
        canvas._plt_connect_mouse_release(self.mouse.moved.emit)
        canvas._plt_connect_mouse_double_click(self.mouse.double_clicked.emit)
        canvas._plt_connect_mouse_double_click(self.mouse.moved.emit)

    def _emit_xlim_changed(self, lim):
        self.x.events.lim.emit(lim)
        self.events.lims.emit(Rect(*lim, *self.y.lim))

    def _emit_ylim_changed(self, lim):
        self.y.events.lim.emit(lim)
        self.events.lims.emit(Rect(*self.x.lim, *lim))

    def _emit_mouse_moved(self, ev):
        """Emit mouse moved event with autoscaling blocked"""
        self.mouse.moved.emit(ev)

    @property
    def mouse_clicked(self):
        warnings.warn(
            "`canvas.events.mouse_clicked` is deprecated. Use `canvas.mouse.clicked` "
            "instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.mouse.clicked

    @property
    def mouse_moved(self):
        warnings.warn(
            "`canvas.events.mouse_moved` is deprecated. Use `canvas.mouse.moved` "
            "instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.mouse.clicked

    @property
    def mouse_double_clicked(self):
        warnings.warn(
            "`canvas.events.mouse_double_clicked` is deprecated. Use "
            "`canvas.mouse.double_clicked` instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.mouse.clicked

    @property
    def native(self) -> Any:
        """Return the native canvas object."""
        return self._canvas()._plt_get_native()

    @property
    def aspect_ratio(self) -> float | None:
        """Aspect ratio of the canvas (None if not locked)."""
        return self._canvas()._plt_get_aspect_ratio()

    @aspect_ratio.setter
    def aspect_ratio(self, ratio: float | None):
        if ratio is not None:
            ratio = float(ratio)
        self._canvas()._plt_set_aspect_ratio(ratio)

    @property
    def mouse_enabled(self) -> bool:
        warnings.warn(
            "`canvas.mouse_enabled` is deprecated. Use `canvas.mouse.enabled` instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.mouse.enabled

    @mouse_enabled.setter
    def mouse_enabled(self, enabled: bool):
        warnings.warn(
            "`canvas.mouse_enabled` is deprecated. Use `canvas.mouse.enabled` instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        self.mouse.enabled = enabled

    def autoscale(
        self,
        xpad: float | tuple[float, float] | None = None,
        ypad: float | tuple[float, float] | None = None,
    ) -> tuple[float, float, float, float]:
        """
        Autoscale the canvas to fit the contents.

        Parameters
        ----------
        xpad : float or (float, float), optional
            Padding in the x direction.
        ypad : float or (float, float), optional
            Padding in the y direction.
        """
        ar = np.stack([layer.bbox_hint() for layer in self.layers], axis=0)
        xmin = np.min(ar[:, 0])
        xmax = np.max(ar[:, 1])
        ymin = np.min(ar[:, 2])
        ymax = np.max(ar[:, 3])
        x0, x1 = self.x.lim
        y0, y1 = self.y.lim
        if np.isnan(xmin):
            xmin = x0
        if np.isnan(xmax):
            xmax = x1
        if np.isnan(ymin):
            ymin = y0
        if np.isnan(ymax):
            ymax = y1
        if xpad is not None:
            xrange = xmax - xmin
            if is_real_number(xpad):
                dx0 = dx1 = xpad * xrange
            else:
                dx0, dx1 = xpad[0] * xrange, xpad[1] * xrange
            xmin -= dx0
            xmax += dx1
        if ypad is not None:
            yrange = ymax - ymin
            if is_real_number(ypad):
                dy0 = dy1 = ypad * yrange
            else:
                dy0, dy1 = ypad[0] * yrange, ypad[1] * yrange
            ymin -= dy0
            ymax += dy1
        small_diff = 1e-6
        if xmax - xmin < small_diff:
            xmin -= 0.05
            xmax += 0.05
        if ymax - ymin < small_diff:
            ymin -= 0.05
            ymax += 0.05
        self.x.lim = xmin, xmax
        self.y.lim = ymin, ymax
        return xmin, xmax, ymin, ymax

    def install_second_x(self, *, palette: ColormapType | None = None) -> Canvas:
        """Create a twin canvas that has a secondary x-axis and shared y-axis."""
        try:
            new = self._canvas()._plt_twiny()
        except AttributeError:
            raise NotImplementedError(
                f"Backend {self._get_backend()} does not support `install_second_x`."
            ) from None
        canvas = Canvas.from_backend(new, palette=palette, backend=self._get_backend())
        canvas._init_canvas()
        return canvas

    def install_second_y(self, *, palette: ColormapType | None = None) -> Canvas:
        """Create a twin canvas that has a secondary y-axis and shared x-axis."""
        try:
            new = self._canvas()._plt_twinx()
        except AttributeError:
            raise NotImplementedError(
                f"Backend {self._get_backend()} does not support `install_second_y`."
            ) from None
        canvas = Canvas.from_backend(new, palette=palette, backend=self._get_backend())
        canvas._init_canvas()
        return canvas

    @overload
    def install_inset(
        self, left: float, right: float, bottom: float, top: float, *,
        palette: ColormapType | None = None
    ) -> Canvas:  # fmt: skip
        ...

    @overload
    def install_inset(
        self, rect: Rect | tuple[float, float, float, float], /, *,
        palette: ColormapType | None = None
    ) -> Canvas:  # fmt: skip
        ...

    def install_inset(self, *args, palette=None, **kwargs) -> Canvas:
        """
        Install a new canvas pointing to an inset of the current canvas.

        >>> canvas.install_inset(left=0.1, right=0.9, bottom=0.1, top=0.9)
        >>> canvas.install_inset([0.1, 0.9, 0.1, 0.9])  # or a sequence
        """
        # normalize input
        if len(args) == 1 and not kwargs:
            rect = args[0]
            if not isinstance(rect, Rect):
                rect = Rect.with_check(*rect)
        else:
            rect = Rect.with_check(*args, **kwargs)
        try:
            new = self._canvas()._plt_inset(rect)
        except AttributeError:
            raise NotImplementedError(
                f"Backend {self._get_backend()} does not support `install_inset`"
            ) from None
        canvas = Canvas.from_backend(new, palette=palette, backend=self._get_backend())
        canvas._init_canvas()
        return canvas

    @property
    def visible(self):
        """Show the canvas."""
        return self._canvas()._plt_get_visible()

    @visible.setter
    def visible(self, visible):
        """Hide the canvas."""
        self._canvas()._plt_set_visible(visible)

    @property
    def lims(self) -> Rect:
        """Return the x/y limits of the canvas."""
        return Rect(*self.x.lim, *self.y.lim)

    @lims.setter
    def lims(self, lims: tuple[float, float, float, float]):
        xmin, xmax, ymin, ymax = lims
        if xmin >= xmax or ymin >= ymax:
            raise ValueError(f"Invalid view rect: {Rect(*lims)}")
        with self.events.lims.blocked():
            self.x.lim = xmin, xmax
            self.y.lim = ymin, ymax
        self.events.lims.emit(Rect(xmin, xmax, ymin, ymax))

    def update_axes(
        self,
        *,
        visible: bool | None = None,
        color: ColorType | None = None,
    ):
        """
        Update axes appearance.

        Parameters
        ----------
        visible : bool, optional
            Whether to show the axes.
        color : color-like, optional
            Color of the axes.
        """
        if visible is not None:
            self.x.ticks.visible = self.y.ticks.visible = visible
        if color is not None:
            self.x.color = self.y.color = color
            self.x.ticks.color = self.y.ticks.color = color
            self.x.label.color = self.y.label.color = color
        return self

    def update_labels(
        self,
        title: str | None = None,
        x: str | None = None,
        y: str | None = None,
    ) -> Self:
        """
        Helper function to update the title, x, and y labels.

        >>> from whitecanvas import new_canvas
        >>> canvas = new_canvas("matplotlib").update_labels("Title", "X", "Y")
        """
        if title is not None:
            self.title.text = title
            self.title.visible = True
        if x is not None:
            self.x.label.text = x
            self.x.label.visible = True
        if y is not None:
            self.y.label.text = y
            self.y.label.visible = True
        return self

    def update_font(
        self,
        size: float | None = None,
        color: ColorType | None = None,
        family: str | None = None,
    ) -> Self:
        """
        Update all the fonts, including the title, x/y labels and x/y tick labels.

        Parameters
        ----------
        size : float, optional
            New font size.
        color : color-like, optional
            New font color.
        family : str, optional
            New font family.
        """
        if size is not None:
            self.title.size = self.x.label.size = self.y.label.size = size
            self.x.ticks.size = self.y.ticks.size = size
        if family is not None:
            self.title.family = self.x.label.family = self.y.label.family = family
            self.x.ticks.family = self.y.ticks.family = family
        if color is not None:
            self.title.color = self.x.label.color = self.y.label.color = color
            self.x.ticks.color = self.y.ticks.color = color
        return self

    def cat(
        self,
        data: _DF,
        x: str | None = None,
        y: str | None = None,
        *,
        update_labels: bool = True,
    ) -> _df.CatPlotter[Self, _DF]:
        """
        Categorize input data for plotting.

        This method provides categorical plotting methods for the input data.
        Methods are very similar to `seaborn` and `plotly.express`.

        Parameters
        ----------
        data : tabular data
            Any categorizable data. Currently, dict, pandas.DataFrame, and
            polars.DataFrame are supported.
        x : str, optional
            Name of the column that will be used for the x-axis. Must be numerical.
        y : str, optional
            Name of the column that will be used for the y-axis. Must be numerical.
        update_labels : bool, default True
            If True, update the x/y labels to the corresponding names.

        Returns
        -------
        CatPlotter
            Plotter object.
        """
        plotter = _df.CatPlotter(self, data, x, y, update_labels=update_labels)
        return plotter

    def cat_x(
        self,
        data: _DF,
        x: str | Sequence[str] | None = None,
        y: str | None = None,
        *,
        update_labels: bool = True,
        numeric_axis: bool = False,
    ) -> _df.XCatPlotter[Self, _DF]:
        """
        Categorize input data for plotting with x-axis as a categorical axis.

        Parameters
        ----------
        data : tabular data
            Any categorizable data. Currently, dict, pandas.DataFrame, and
            polars.DataFrame are supported.
        x : str or sequence of str, optional
            Name of the column(s) that will be used for the x-axis. Must be categorical.
        y : str, optional
            Name of the column that will be used for the y-axis. Must be numerical.
        update_labels : bool, default True
            If True, update the x/y labels to the corresponding names.
        numeric_axis : bool, default False
            If True, the x-axis will be treated as a numerical axis. For example, if
            categories are [2, 4, 8], the x coordinates will be mapped to [0, 1, 2] by
            default, but if this option is True, the x coordinates will be [2, 4, 8].

        Returns
        -------
        XCatPlotter
            Plotter object.
        """
        return _df.XCatPlotter(self, data, x, y, update_labels, numeric=numeric_axis)

    def cat_y(
        self,
        data: _DF,
        x: str | None = None,
        y: str | Sequence[str] | None = None,
        *,
        update_labels: bool = True,
        numeric_axis: bool = False,
    ) -> _df.YCatPlotter[Self, _DF]:
        """
        Categorize input data for plotting with y-axis as a categorical axis.

        Parameters
        ----------
        data : tabular data
            Any categorizable data. Currently, dict, pandas.DataFrame, and
            polars.DataFrame are supported.
        x : str, optional
            Name of the column that will be used for the x-axis. Must be numerical.
        y : str or sequence of str, optional
            Name of the column(s) that will be used for the y-axis. Must be categorical.
        update_labels : bool, default True
            If True, update the x/y labels to the corresponding names.
        numeric_axis : bool, default False
            If True, the x-axis will be treated as a numerical axis. For example, if
            categories are [2, 4, 8], the y coordinates will be mapped to [0, 1, 2] by
            default, but if this option is True, the y coordinates will be [2, 4, 8].

        Returns
        -------
        YCatPlotter
            Plotter object
        """
        return _df.YCatPlotter(self, data, y, x, update_labels, numeric=numeric_axis)

    def cat_xy(
        self,
        data: _DF,
        x: str | Sequence[str],
        y: str | Sequence[str],
        *,
        update_labels: bool = True,
    ) -> _df.XYCatPlotter[Self, _DF]:
        """
        Categorize input data for plotting with both axes as categorical.

        Parameters
        ----------
        data : tabular data
            Any categorizable data. Currently, dict, pandas.DataFrame, and
            polars.DataFrame are supported.
        x : str or sequence of str, optional
            Name of the column(s) that will be used for the x-axis. Must be categorical.
        y : str or sequence of str, optional
            Name of the column(s) that will be used for the y-axis. Must be categorical.
        update_labels : bool, default True
            If True, update the x/y labels to the corresponding names.

        Returns
        -------
        XYCatPlotter
            Plotter object
        """
        return _df.XYCatPlotter(self, data, x, y, update_labels)

    def stack_over(self, layer: _L0) -> StackOverPlotter[Self, _L0]:
        """
        Stack new data over the existing layer.

        For example following code

        >>> bars_0 = canvas.add_bars(x, y0)
        >>> bars_1 = canvas.stack_over(bars_0).add(y1)
        >>> bars_2 = canvas.stack_over(bars_1).add(y2)

        will result in a bar plot like this

        ```
         ┌───┐
         ├───│┌───┐
         │   │├───│
         ├───│├───│
        ─┴───┴┴───┴─
        ```
        """
        if not isinstance(layer, (_l.Bars, _l.Band, _lg.StemPlot, _lg.LabeledBars)):
            raise TypeError(
                f"Only Bars, StemPlot and Band are supported as an input, "
                f"got {type(layer)!r}."
            )
        return StackOverPlotter(self, layer)

    # TODO
    # def annotate(self, layer, at: int):
    #     ...

    def between(self, l0, l1) -> BetweenPlotter[Self]:
        return BetweenPlotter(self, l0, l1)

    def fit(self, layer: _l.DataBoundLayer[_P]) -> FitPlotter[Self, _P]:
        """The fit plotter namespace."""
        return FitPlotter(self, layer)

    def add_legend(
        self,
        layers: Sequence[str | _l.Layer] | None = None,
        *,
        location: Location | LocationStr = "top_right",
        title: str | None = None,
    ):
        """
        Add legend items to the canvas.

        Parameters
        ----------
        layers : sequence of layer or str, optional
            Which item to be added to the legend. If str is given, it will be converted
            into a legend title label.
        location : LegendLocation, default "top_right"
            Location of the legend. Can be combination of "top", "bottom", "left",
            "right" and "center" (e.g., "top_left", "center_right").

            ```
                   (2) left  center right
                         v     v     v
              (1)     ┌─────────────────┐
               top -> │                 │
            center -> │     canvas      │
            bottom -> │                 │
                      └─────────────────┘
            ```

            Some backends also support adding legend outside the canvas. Following
            strings suffixed with "_side" can be used in combination with those strings
            above (e.g., "bottom_side_rigth", "right_side_top").

            ```
               top_side -> ┌────────┐
                        ┌──┼────────┼──┐
            left_side ->│  │ canvas │  │<- right_side
                        └──┼────────┼──┘
            bottom_side -> └────────┘
            ```
        title : str, optional
            If given, title label will be added as the first legend item.
        """
        if layers is None:
            layers = list(self.layers)
        if title is not None:
            layers = [title, *layers]
        location = Location.parse(location)

        items = list[tuple[str, _legend.LegendItem]]()
        for layer in layers:
            if isinstance(layer, str):
                items.append((layer, _legend.TitleItem()))
            elif isinstance(layer, _l.Layer):
                items.append((layer.name, layer._as_legend_item()))
            else:
                raise TypeError(f"Expected a list of layer or str, got {type(layer)}.")
        self._canvas()._plt_make_legend(items, location)

    @overload
    def add_line(
        self, ydata: ArrayLike1D, *, name: str | None = None,
        color: ColorType | None = None, width: float = 1.0,
        style: LineStyle | str = LineStyle.SOLID, alpha: float = 1.0,
        antialias: bool = True,
    ) -> _l.Line:  # fmt: skip
        ...

    @overload
    def add_line(
        self, xdata: ArrayLike1D, ydata: ArrayLike1D, *, name: str | None = None,
        color: ColorType | None = None, width: float | None = None,
        style: LineStyle | str | None = None, alpha: float = 1.0,
        antialias: bool = True,
    ) -> _l.Line:  # fmt: skip
        ...

    @overload
    def add_line(
        self, xdata: ArrayLike1D, ydata: Callable[[ArrayLike1D], ArrayLike1D], *,
        name: str | None = None, color: ColorType | None = None,
        width: float | None = None, style: LineStyle | str | None = None,
        alpha: float = 1.0, antialias: bool = True,
    ) -> _l.Line:  # fmt: skip
        ...

    def add_line(
        self,
        *args,
        name=None,
        color=None,
        width=None,
        style=None,
        alpha=1.0,
        antialias=True,
    ):
        """
        Add a Line layer to the canvas.

        >>> canvas.add_line(y, ...)
        >>> canvas.add_line(x, y, ...)

        Parameters
        ----------
        name : str, optional
            Name of the layer.
        color : color-like, optional
            Color of the bars.
        width : float, optional
            Line width. Use the theme default if not specified.
        style : str or LineStyle, optional
            Line style. Use the theme default if not specified.
        alpha : float, default 1.0
            Alpha channel of the line.
        antialias : bool, default True
            Antialiasing of the line.

        Returns
        -------
        Line
            The line layer.
        """
        xdata, ydata = normalize_xy(*args)
        name = self._coerce_name(name)
        color = self._generate_colors(color)
        width = theme._default("line.width", width)
        style = theme._default("line.style", style)
        layer = _l.Line(
            xdata, ydata, name=name, color=color, width=width, style=style,
            alpha=alpha, antialias=antialias, backend=self._get_backend(),
        )  # fmt: skip
        return self.add_layer(layer)

    @overload
    def add_markers(
        self, xdata: ArrayLike1D, ydata: ArrayLike1D, *,
        name: str | None = None, symbol: Symbol | str | None = None,
        size: float | None = None, color: ColorType | None = None, alpha: float = 1.0,
        hatch: str | Hatch | None = None,
    ) -> _l.Markers[_mixin.ConstFace, _mixin.ConstEdge, float]:  # fmt: skip
        ...

    @overload
    def add_markers(
        self, ydata: ArrayLike1D, *,
        name: str | None = None, symbol: Symbol | str | None = None,
        size: float | None = None, color: ColorType | None = None, alpha: float = 1.0,
        hatch: str | Hatch | None = None,
    ) -> _l.Markers[_mixin.ConstFace, _mixin.ConstEdge, float]:  # fmt: skip
        ...

    def add_markers(
        self,
        *args,
        name=None,
        symbol=None,
        size=None,
        color=None,
        alpha=1.0,
        hatch=None,
    ):
        """
        Add markers (scatter plot).

        >>> canvas.add_markers(x, y)  # standard usage
        >>> canvas.add_markers(y)  # use 0, 1, ... for the x values

        Parameters
        ----------
        name : str, optional
            Name of the layer.
        symbol : str or Symbol, optional
            Marker symbols. Use the theme default if not specified.
        size : float, optional
            Marker size. Use the theme default if not specified.
        color : color-like, optional
            Color of the marker faces.
        alpha : float, default 1.0
            Alpha channel of the marker faces.
        hatch : str or FacePattern, optional
            Pattern of the marker faces. Use the theme default if not specified.

        Returns
        -------
        Markers
            The markers layer.
        """
        xdata, ydata = normalize_xy(*args)
        name = self._coerce_name(name)
        color = self._generate_colors(color)
        symbol = theme._default("markers.symbol", symbol)
        size = theme._default("markers.size", size)
        hatch = theme._default("markers.hatch", hatch)
        layer = _l.Markers(
            xdata, ydata, name=name, symbol=symbol, size=size, color=color,
            alpha=alpha, hatch=hatch, backend=self._get_backend(),
        )  # fmt: skip
        return self.add_layer(layer)

    @overload
    def add_step(
        self, ydata: ArrayLike1D, *, name: str | None = None,
        where: StepStyleStr | StepStyle = "pre", color: ColorType | None = None,
        width: float | None = None, style: LineStyle | str | None = None,
        alpha: float = 1.0, orient: OrientationLike = "horizontal",
        antialias: bool = True,
    ) -> _l.LineStep:  # fmt: skip
        ...

    @overload
    def add_step(
        self, xdata: ArrayLike1D, ydata: ArrayLike1D, *, name: str | None = None,
        where: StepStyleStr | StepStyle = "pre", color: ColorType | None = None,
        width: float | None = None, style: LineStyle | str | None = None,
        alpha: float = 1.0, orient: OrientationLike = "horizontal",
        antialias: bool = True,
    ) -> _l.LineStep:  # fmt: skip
        ...

    def add_step(
        self,
        *args,
        name=None,
        where="pre",
        color=None,
        width=None,
        style=None,
        alpha=1.0,
        antialias=True,
    ):
        """
        Add a step plot to the canvas.

        >>> canvas.add_step(y, ...)
        >>> canvas.add_step(x, y, ...)

        Parameters
        ----------
        name : str, optional
            Name of the layer.
        where : str or StepStyle, default "pre"
            Where the step should be placed.
        color : color-like, optional
            Color of the steps.
        width : float, optional
            Line width. Use the theme default if not specified.
        style : str or LineStyle, optional
            Line style. Use the theme default if not specified.
        alpha : float, default 1.0
            Alpha channel of the line.
        antialias : bool, default True
            Antialiasing of the line.

        Returns
        -------
        LineStep
            The line-step layer.
        """
        xdata, ydata = normalize_xy(*args)
        name = self._coerce_name(name)
        color = self._generate_colors(color)
        width = theme._default("line.width", width)
        style = theme._default("line.style", style)
        layer = _l.LineStep(
            xdata, ydata, name=name, color=color, width=width, style=style, where=where,
            alpha=alpha, antialias=antialias, backend=self._get_backend()
        )  # fmt: skip
        return self.add_layer(layer)

    @overload
    def add_bars(
        self, center: ArrayLike1D, height: ArrayLike1D, *,
        bottom: ArrayLike1D | None = None, name=None,
        orient: OrientationLike = "vertical", extent: float | None = None,
        color: ColorType | None = None, alpha: float = 1.0,
        hatch: str | Hatch | None = None,
    ) -> _l.Bars[_mixin.ConstFace, _mixin.ConstEdge]:  # fmt: skip
        ...

    @overload
    def add_bars(
        self, height: ArrayLike1D, *, bottom: ArrayLike1D | None = None,
        name=None, orient: OrientationLike = "vertical",
        extent: float | None = None, color: ColorType | None = None,
        alpha: float = 1.0, hatch: str | Hatch | None = None,
    ) -> _l.Bars[_mixin.ConstFace, _mixin.ConstEdge]:  # fmt: skip
        ...

    def add_bars(
        self,
        *args,
        bottom=None,
        name=None,
        orient="vertical",
        extent=None,
        color=None,
        alpha=1.0,
        hatch=None,
    ):
        """
        Add a bar plot.

        >>> canvas.add_bars(x, heights)  # standard usage
        >>> canvas.add_bars(heights)  # use 0, 1, ... for the x values
        >>> canvas.add_bars(..., orient="horizontal")  # horizontal bars

        Parameters
        ----------
        bottom : float or array-like, optional
            Bottom level of the bars.
        name : str, optional
            Name of the layer.
        orient : str or Orientation, default "vertical"
            Orientation of the bars.
        extent : float, default 0.8
            Bar width in the canvas coordinate
        color : color-like, optional
            Color of the bars.
        alpha : float, default 1.0
            Alpha channel of the bars.
        hatch : str or FacePattern, default FacePattern.SOLID
            Pattern of the bar faces.

        Returns
        -------
        Bars
            The bars layer.
        """
        center, height = normalize_xy(*args)
        if bottom is not None:
            bottom = as_array_1d(bottom)
            if bottom.shape != height.shape:
                raise ValueError("Expected bottom to have the same shape as height")
        name = self._coerce_name(name)
        color = self._generate_colors(color)
        extent = theme._default("bars.extent", extent)
        hatch = theme._default("bars.hatch", hatch)
        layer = _l.Bars(
            center, height, bottom, extent=extent, name=name, orient=orient,
            color=color, alpha=alpha, hatch=hatch, backend=self._get_backend(),
        )  # fmt: skip
        return self.add_layer(layer)

    def add_hist(
        self,
        data: ArrayLike1D,
        *,
        bins: HistBinType = "auto",
        limits: tuple[float, float] | None = None,
        name: str | None = None,
        shape: Literal["step", "polygon", "bars"] = "bars",
        kind: Literal["count", "density", "frequency", "percent"] = "count",
        orient: OrientationLike = "vertical",
        color: ColorType | None = None,
        width: float | None = None,
        style: LineStyle | str | None = None,
    ) -> _lg.Histogram:
        """
        Add data as a histogram.

        >>> canvas.add_hist(np.random.normal(size=100), bins=12)

        Parameters
        ----------
        data : array-like
            1D Array of data.
        bins : int or 1D array-like, default "auto"
            Bins of the histogram. This parameter will directly be passed
            to `np.histogram`.
        limits : (float, float), optional
            Limits in which histogram will be built. This parameter will equivalent to
            the `range` paraneter of `np.histogram`.
        name : str, optional
            Name of the layer.
        shape : {"step", "polygon", "bars"}, default "bars"
            Shape of the histogram. This parameter defines how to convert the data into
            the line nodes.
        kind : {"count", "density", "probability", "frequency", "percent"}, optional
            Kind of the histogram.
        orient : str or Orientation, default "vertical"
            Orientation of the bars.
        color : color-like, optional
            Color of the bars.

        Returns
        -------
        Bars
            The bars layer that represents the histogram.
        """
        name = self._coerce_name(name)
        color = self._generate_colors(color)
        width = theme._default("line.width", width)
        style = theme._default("line.style", style)
        layer = _lg.Histogram.from_array(
            data, bins=bins, limits=limits, shape=shape, kind=kind, name=name,
            color=color, width=width, style=style, orient=orient,
            backend=self._get_backend(),
        )  # fmt: skip
        return self.add_layer(layer)

    def add_hist2d(
        self,
        x: ArrayLike1D,
        y: ArrayLike1D,
        *,
        cmap: ColormapType = "inferno",
        name: str | None = None,
        bins: HistBinType | tuple[HistBinType, HistBinType] = "auto",
        rangex: tuple[float, float] | None = None,
        rangey: tuple[float, float] | None = None,
        density: bool = False,
    ) -> _l.Image:
        """
        Add a 2D histogram of given X/Y data.

        >>> x = np.random.normal(size=100)
        >>> y = np.random.normal(size=200)
        >>> canvas.add_hist2d(x, y)

        Note that unlike `add_image()` method, this method does not lock the aspect
        ratio and flip the canvas by default.

        Parameters
        ----------
        x : array-like
            1D Array of X data.
        y : array-like
            1D Array of Y data.
        cmap : ColormapType, default "gray"
            Colormap used for the image.
        name : str, optional
            Name of the layer.
        bins : int or tuple[int, int], optional
            Bins of the histogram of X/Y dimension respectively. If an integer is given,
            it will be used for both dimensions.
        rangex : (float, float), optional
            Range of x values in which histogram will be built.
        rangey : (float, float), optional
            Range of y values in which histogram will be built.
        density : bool, default False
            If True, values of the histogram will be normalized so that the total
            intensity of the histogram will be 1.

        Returns
        -------
        Image
            Image layer representing the 2D histogram.
        """
        layer = _l.Image.build_hist(
            x, y, bins=bins, range=(rangex, rangey), density=density, name=name,
            cmap=cmap, backend=self._get_backend(),
        )  # fmt: skip
        return self.add_layer(layer)

    def add_rects(
        self,
        coords: ArrayLike,
        *,
        name=None,
        color: ColorType | None = None,
        alpha: float = 1.0,
        hatch: str | Hatch | None = None,
    ) -> _l.Rects[_mixin.ConstFace, _mixin.ConstEdge]:
        name = self._coerce_name(name)
        color = self._generate_colors(color)
        hatch = theme._default("bars.hatch", hatch)
        layer = _l.Rects(
            coords, name=name, color=color, alpha=alpha, hatch=hatch,
            backend=self._get_backend()
        )  # fmt: skip
        return self.add_layer(layer)

    def add_cdf(
        self,
        data: ArrayLike1D,
        *,
        name: str | None = None,
        orient: OrientationLike = "vertical",
        color: ColorType | None = None,
        width: float | None = None,
        style: LineStyle | str | None = None,
        alpha: float = 1.0,
        antialias: bool = True,
    ) -> _l.Line:
        """
        Add a empirical cumulative distribution function (CDF) plot.

        >>> canvas.add_cdf(np.random.normal(size=100))

        Parameters
        ----------
        data : array-like
            1D Array of data.
        name : str, optional
            Name of the layer.
        orient : str or Orientation, default "vertical"
            Orientation of the bars.
        color : color-like, optional
            Color of the bars.
        width : float, optional
            Line width. Use the theme default if not specified.
        style : str or LineStyle, optional
            Line style. Use the theme default if not specified.
        alpha : float, default 1.0
            Alpha channel of the line.
        antialias : bool, default True
            Antialiasing of the line.

        Returns
        -------
        Line
            The line layer that represents the CDF.
        """
        name = self._coerce_name(name)
        color = self._generate_colors(color)
        width = theme._default("line.width", width)
        style = theme._default("line.style", style)
        layer = _l.Line.build_cdf(
            data, orient=orient, name=name, color=color, width=width, style=style,
            alpha=alpha, antialias=antialias, backend=self._get_backend(),
        )  # fmt: skip
        return self.add_layer(layer)

    def add_spans(
        self,
        spans: ArrayLike,
        *,
        name: str | None = None,
        orient: OrientationLike = "vertical",
        color: ColorType = "blue",
        alpha: float = 0.4,
        hatch: str | Hatch = Hatch.SOLID,
    ) -> _l.Spans:
        """
        Add spans that extends infinitely.

        >>> canvas.add_spans([[5, 10], [15, 20]])

           |::::|     |::::|
           |::::|     |::::|
        ───5────10────15───20─────>
           |::::|     |::::|
           |::::|     |::::|

        Parameters
        ----------
        spans : (N, 2) array-like
            Array that contains the start and end points of the spans.
        name : str, optional
            Name of the layer.
        orient : str or Orientation, default "vertical"
            Orientation of the bars.
        color : color-like, optional
            Color of the bars.
        alpha : float, default 0.4
            Alpha channel of the bars.
        hatch : str or FacePattern, default FacePattern.SOLID
            Pattern of the bar faces.

        Returns
        -------
        Spans
            The spans layer.
        """
        name = self._coerce_name(name)
        color = self._generate_colors(color)
        layer = _l.Spans(
            spans, name=name, orient=orient, color=color, alpha=alpha,
            hatch=hatch, backend=self._get_backend(),
        )  # fmt: skip
        return self.add_layer(layer)

    def add_vectors(
        self,
        x: ArrayLike1D,
        y: ArrayLike1D,
        vx: ArrayLike1D,
        vy: ArrayLike1D,
        *,
        name: str | None = None,
        color: ColorType | None = None,
        width: float | None = None,
        style: LineStyle | str | None = None,
        alpha: float = 1.0,
        antialias: bool = True,
    ) -> _l.Vectors:
        """
        Add a vector field to the canvas.

        >>> canvas.add_vectors(x, y, vx, vy)

        Parameters
        ----------
        x : array-like
            X coordinates of the vectors.
        y : array-like
            Y coordinates of the vectors.
        vx : array-like
            X components of the vectors.
        vy : array-like
            Y components of the vectors.
        name : str, optional
            Name of the layer.
        color : color-like, optional
            Color of the bars.
        width : float, optional
            Line width. Use the theme default if not specified.
        style : str or LineStyle, optional
            Line style. Use the theme default if not specified.
        alpha : float, default 1.0
            Alpha channel of the line.
        antialias : bool, default True
            Antialiasing of the line.

        Returns
        -------
        Vectors
            The vectors layer.
        """
        name = self._coerce_name(name)
        color = self._generate_colors(color)
        width = theme._default("line.width", width)
        style = theme._default("line.style", style)
        layer = _l.Vectors(
            as_array_1d(x, dtype=np.float32), as_array_1d(y, dtype=np.float32),
            as_array_1d(vx, dtype=np.float32), as_array_1d(vy, dtype=np.float32),
            name=name, color=color, width=width, style=style,
            alpha=alpha, antialias=antialias, backend=self._get_backend(),
        )  # fmt: skip
        return self.add_layer(layer)

    def add_infline(
        self,
        pos: tuple[float, float] = (0, 0),
        angle: float = 0.0,
        *,
        name: str | None = None,
        color: ColorType | None = None,
        width: float | None = None,
        style: LineStyle | str | None = None,
        alpha: float = 1.0,
        antialias: bool = True,
    ) -> _l.InfLine:
        """
        Add an infinitely long line to the canvas.

        >>> canvas.add_infline((0, 0), 45)  # y = x
        >>> canvas.add_infline((1, 0), 90)  # x = 1
        >>> canvas.add_infline((0, -1), 0)  # y = -1

        Parameters
        ----------
        pos : (float, float), default (0, 0)
            One of the points this line passes.
        angle : float, default 0.0
            Angle of the line in degree, defined by the counter-clockwise
            rotation from the x axis.
        name : str, optional
            Name of the layer.
        color : color-like, optional
            Color of the bars.
        width : float, optional
            Line width. Use the theme default if not specified.
        style : str or LineStyle, optional
            Line style. Use the theme default if not specified.
        alpha : float, default 1.0
            Alpha channel of the line.
        antialias : bool, default True
            Antialiasing of the line.

        Returns
        -------
        InfLine
            The infline layer.
        """
        name = self._coerce_name(name)
        color = self._generate_colors(color)
        width = theme._default("line.width", width)
        style = theme._default("line.style", style)
        layer = _l.InfLine(
            pos, angle, name=name, color=color, alpha=alpha,
            width=width, style=style, antialias=antialias,
            backend=self._get_backend(),
        )  # fmt: skip
        return self.add_layer(layer)

    def add_infcurve(
        self,
        model: Callable[Concatenate[Any, _P], Any],
        *,
        bounds: tuple[float, float] = (-float("inf"), float("inf")),
        name: str | None = None,
        color: ColorType | None = None,
        width: float | None = None,
        style: str | LineStyle | None = None,
        alpha: float = 1.0,
        antialias: bool = True,
    ) -> _l.InfCurve[_P]:
        """
        Add an infinite curve to the canvas.

        >>> canvas.add_infcurve(lambda x: x ** 2)  # parabola
        >>> canvas.add_infcurve(lambda x, a: np.sin(a*x)).update_params(2)  # parametric

        Parameters
        ----------
        model : callable
            The model function. The first argument must be the x coordinates. Same
            signature as `scipy.optimize.curve_fit`.
        bounds : (float, float), default (-inf, inf)
            Lower and upper bounds that the function is defined.
        name : str, optional
            Name of the layer.
        color : color-like, optional
            Color of the bars.
        width : float, optional
            Line width. Use the theme default if not specified.
        style : str or LineStyle, optional
            Line style. Use the theme default if not specified.
        alpha : float, default 1.0
            Alpha channel of the line.
        antialias : bool, default True
            Antialiasing of the line.

        Returns
        -------
        InfCurve
            The infcurve layer.
        """
        name = self._coerce_name(name)
        color = self._generate_colors(color)
        width = theme._default("line.width", width)
        style = theme._default("line.style", style)
        layer = _l.InfCurve(
            model, bounds=bounds, name=name, color=color, width=width, alpha=alpha,
            style=style, antialias=antialias, backend=self._get_backend(),
        )  # fmt: skip
        return self.add_layer(layer)

    def add_hline(
        self,
        y: float,
        *,
        name: str | None = None,
        color: ColorType | None = None,
        width: float | None = None,
        style: LineStyle | str = LineStyle.SOLID,
        alpha: float = 1.0,
        antialias: bool = True,
    ) -> _l.InfLine:
        """
        Add a infinite horizontal line to the canvas.

        Parameters
        ----------
        y : float
            Y coordinate of the line.
        name : str, optional
            Name of the layer.
        color : color-like, optional
            Color of the bars.
        width : float, optional
            Line width. Use the theme default if not specified.
        style : str or LineStyle, optional
            Line style. Use the theme default if not specified.
        alpha : float, default 1.0
            Alpha channel of the line.
        antialias : bool, default True
            Antialiasing of the line.

        Returns
        -------
        InfLine
            The infline layer.
        """
        return self.add_infline(
            (0, y), 0, name=name, color=color, width=width, style=style, alpha=alpha,
            antialias=antialias
        )  # fmt: skip

    def add_vline(
        self,
        x: float,
        *,
        name: str | None = None,
        color: ColorType | None = None,
        width: float | None = None,
        style: LineStyle | str = LineStyle.SOLID,
        alpha: float = 1.0,
        antialias: bool = True,
    ) -> _l.InfLine:
        """
        Add a infinite vertical line to the canvas.

        Parameters
        ----------
        x : float
            X coordinate of the line.
        name : str, optional
            Name of the layer.
        color : color-like, optional
            Color of the bars.
        width : float, optional
            Line width. Use the theme default if not specified.
        style : str or LineStyle, optional
            Line style. Use the theme default if not specified.
        alpha : float, default 1.0
            Alpha channel of the line.
        antialias : bool, default True
            Antialiasing of the line.

        Returns
        -------
        InfLine
            The infline layer.
        """
        return self.add_infline(
            (x, 0), 90, name=name, color=color, width=width, style=style, alpha=alpha,
            antialias=antialias,
        )  # fmt: skip

    def add_band(
        self,
        xdata: ArrayLike1D,
        ylow: ArrayLike1D,
        yhigh: ArrayLike1D,
        *,
        name: str | None = None,
        orient: OrientationLike = "vertical",
        color: ColorType | None = None,
        alpha: float = 1.0,
        hatch: str | Hatch = Hatch.SOLID,
    ) -> _l.Band:
        """
        Add a band (fill-between) layer to the canvas.

        Parameters
        ----------
        xdata : array-like
            X coordinates of the band.
        ylow : array-like
            Either lower or upper y coordinates of the band.
        yhigh : array-like
            The other y coordinates of the band.
        name : str, optional
            Name of the layer, by default None
        orient : str, Orientation, default "vertical"
            Orientation of the band. If vertical, band will be filled between
            vertical orientation.,
        color : color-like, default None
            Color of the band face.,
        alpha : float, default 1.0
            Alpha channel of the band face.
        hatch : str, FacePattern, default FacePattern.SOLID
            Hatch of the band face.

        Returns
        -------
        Band
            The band layer.
        """
        name = self._coerce_name(name)
        color = self._generate_colors(color)
        layer = _l.Band(
            xdata, ylow, yhigh, name=name, orient=orient, color=color,
            alpha=alpha, hatch=hatch, backend=self._get_backend(),
        )  # fmt: skip
        return self.add_layer(layer)

    def add_errorbars(
        self,
        xdata: ArrayLike1D,
        ylow: ArrayLike1D,
        yhigh: ArrayLike1D,
        *,
        name: str | None = None,
        orient: OrientationLike = "vertical",
        color: ColorType | None = None,
        width: float | None = None,
        style: LineStyle | str | None = None,
        alpha: float = 1.0,
        antialias: bool = False,
        capsize: float = 0.0,
    ) -> _l.Errorbars:
        """
        Add parallel lines as errorbars.

        Parameters
        ----------
        xdata : array-like
            X coordinates of the errorbars.
        ylow : array-like
            Lower bound of the errorbars.
        yhigh : array-like
            Upper bound of the errorbars.
        name : str, optional
            Name of the layer.
        orient : str or Orientation, default "vertical"
            Orientation of the errorbars. If vertical, errorbars will be parallel
            to the y axis.
        color : color-like, optional
            Color of the bars.
        width : float, optional
            Line width. Use the theme default if not specified.
        style : str or LineStyle, optional
            Line style. Use the theme default if not specified.
        alpha : float, default 1.0
            Alpha channel of the line.
        antialias : bool, default True
            Antialiasing of the line.
        capsize : float, default 0.0
            Size of the caps of the error indicators

        Returns
        -------
        Errorbars
            The errorbars layer.
        """
        name = self._coerce_name(name)
        color = self._generate_colors(color)
        width = theme._default("line.width", width)
        style = theme._default("line.style", style)
        layer = _l.Errorbars(
            xdata, ylow, yhigh, name=name, color=color, width=width,
            style=style, antialias=antialias, capsize=capsize, alpha=alpha,
            orient=orient, backend=self._get_backend(),
        )  # fmt: skip
        return self.add_layer(layer)

    def add_rug(
        self,
        events: ArrayLike1D,
        *,
        low: float = 0.0,
        high: float = 1.0,
        name: str | None = None,
        orient: OrientationLike = "vertical",
        color: ColorType = "black",
        width: float = 1.0,
        style: LineStyle | str = LineStyle.SOLID,
        antialias: bool = True,
        alpha: float = 1.0,
    ) -> _l.Rug:
        """
        Add input data as a rug plot.

        >>> canvas.add_rug([2, 4, 5, 8, 11])

        ```
          │ ││  │   │
        ──┴─┴┴──┴───┴──> x
          2 45  8   11
        ```

        Parameters
        ----------
        events : array-like
            A 1D array of events.
        low : float, default 0.0
            The lower bound of the rug lines.
        high : float, default 1.0
            The upper bound of the rug lines.
        name : str, optional
            Name of the layer.
        orient : str or Orientation, default "vertical"
            Orientation of the errorbars. If vertical, rug lines will be parallel
            to the y axis.
        color : color-like, optional
            Color of the bars.
        width : float, default 1.0
            Line width.
        style : str or LineStyle, default LineStyle.SOLID
            Line style.
        alpha : float, default 1.0
            Alpha channel of the line.
        antialias : bool, default True
            Antialiasing of the line.

        Returns
        -------
        Rug
            The rug layer.
        """
        name = self._coerce_name(name)
        color = self._generate_colors(color)
        layer = _l.Rug(
            events, low=low, high=high, name=name, color=color, alpha=alpha,
            width=width, style=style, antialias=antialias, orient=orient,
            backend=self._get_backend(),
        )  # fmt: skip
        return self.add_layer(layer)

    def add_kde(
        self,
        data: ArrayLike1D,
        *,
        bottom: float = 0.0,
        name: str | None = None,
        orient: OrientationLike = "vertical",
        band_width: KdeBandWidthType = "scott",
        color: ColorType | None = None,
        width: float | None = None,
        style: LineStyle | str | None = None,
    ) -> _lg.Kde:
        """
        Add data as a band layer representing kernel density estimation (KDE).

        Parameters
        ----------
        data : array-like
            1D data to calculate the KDE.
        bottom : float, default 0.0
            Scalar value that define the height of the bottom line.
        name : str, optional
            Name of the layer, by default None
        orient : str, Orientation, default "vertical"
            Orientation of the KDE.
        band_width : float or str, default "scott"
            Method to calculate the estimator bandwidth.
        color : color-like, default None
            Color of the band face.
        width : float, optional
            Line width of the outline.
        style : str or LineStyle, optional
            Line style of the outline.

        Returns
        -------
        Kde
            The KDE layer.
        """
        name = self._coerce_name(name)
        color = self._generate_colors(color)
        width = theme._default("line.width", width)
        style = theme._default("line.style", style)

        layer = _lg.Kde.from_array(
            data, bottom=bottom, scale=1, band_width=band_width, name=name,
            orient=orient, color=color, width=width, style=style,
            backend=self._get_backend(),
        )  # fmt: skip
        return self.add_layer(layer)

    @overload
    def add_text(
        self, x: ArrayLike1D, y: ArrayLike1D, string: list[str], *,
        color: ColorType = "black", size: float = 12, rotation: float = 0.0,
        anchor: str | Alignment = Alignment.BOTTOM_LEFT, family: str | None = None,
    ) -> _l.Texts[_mixin.ConstFace, _mixin.ConstEdge, _mixin.ConstFont]:  # fmt: skip
        ...

    @overload
    def add_text(
        self, x: float, y: float, string: str, *, color: ColorType = "black",
        size: float = 12, rotation: float = 0.0,
        anchor: str | Alignment = Alignment.BOTTOM_LEFT, family: str | None = None,
    ) -> _l.Texts[_mixin.ConstFace, _mixin.ConstEdge, _mixin.ConstFont]:  # fmt: skip
        ...

    def add_text(
        self,
        x,
        y,
        string,
        *,
        color="black",
        size=12,
        rotation=0.0,
        anchor=Alignment.BOTTOM_LEFT,
        family=None,
    ):
        """
        Add a text layer to the canvas.

        >>> canvas.add_text([0, 0], [1, 1], ["text-0", "text-1])
        >>> canvas.add_text(...).with_face(color="red")  # with background
        >>> canvas.add_text(...).with_edge(color="red")  # with outline

        Parameters
        ----------
        x : float or array-like
            X position(s) of the text.
        y : float or array-like
            Y position(s) of the text.
        string : str or list[str]
            Text string to display.
        color : ColorType, optional
            Color of the text string.
        size : float, default 12
            Point size of the text.
        rotation : float, default 0.0
            Rotation angle of the text in degrees.
        anchor : str or Alignment, default Alignment.BOTTOM_LEFT
            Anchor position of the text. The anchor position will be the coordinate
            given by (x, y).
        family : str, optional
            Font family of the text.

        Returns
        -------
        Texts
            The text layer.
        """
        if is_real_number(x) and is_real_number(y) and isinstance(string, str):
            x, y, string = [x], [y], [string]
        x_, y_ = normalize_xy(x, y)
        if isinstance(string, str):
            string = [string] * x_.size
        elif len(string) != x_.size:
            raise ValueError("Expected string to have the same size as x/y")
        layer = _l.Texts(
            x_, y_, string, color=color, size=size, rotation=rotation, anchor=anchor,
            family=family, backend=self._get_backend(),
        )  # fmt: skip
        return self.add_layer(layer)

    def add_image(
        self,
        image: ArrayLike,
        *,
        name: str | None = None,
        cmap: ColormapType = "gray",
        clim: tuple[float | None, float | None] | None = None,
        flip_canvas: bool = True,
        lock_aspect: bool = True,
    ) -> _l.Image:
        """
        Add an image layer to the canvas.

        This method automatically flips the image vertically by default. `add_heatmap`
        does the similar thing with slightly different default settings.

        Parameters
        ----------
        image : ArrayLike
            Image data. Must be 2D or 3D array. If 3D, the last dimension must be
            RGB(A). Note that the first dimension is the vertical axis.
        cmap : ColormapType, default "gray"
            Colormap used for the image.
        clim : (float or None, float or None) or None
            Contrast limits. If None, the limits are automatically determined by min and
            max of the data. You can also pass None separately to either limit to use
            the default behavior.
        flip_canvas : bool, default True
            If True, flip the canvas vertically so that the image looks normal.
        lock_aspect : bool, default True
            If True, lock the aspect ratio of the canvas to 1:1.

        Returns
        -------
        Image
            The image layer.
        """
        layer = _l.Image(
            image, name=name, cmap=cmap, clim=clim, backend=self._get_backend()
        )
        self.add_layer(layer)
        if flip_canvas and not self.y.flipped:
            self.y.flipped = True
        if lock_aspect:
            self.aspect_ratio = 1.0
        return layer

    def add_heatmap(
        self,
        image: ArrayLike,
        *,
        name: str | None = None,
        cmap: ColormapType = "inferno",
        clim: tuple[float | None, float | None] | None = None,
        flip_canvas: bool = False,
    ) -> _l.Image:
        """
        Add an image layer to the canvas as a heatmap.

        Use `add_image` to add the layer as an image.

        Parameters
        ----------
        image : ArrayLike
            Image data. Must be 2D or 3D array. If 3D, the last dimension must be
            RGB(A). Note that the first dimension is the vertical axis.
        cmap : ColormapType, default "gray"
            Colormap used for the image.
        clim : (float or None, float or None) or None
            Contrast limits. If None, the limits are automatically determined by min and
            max of the data. You can also pass None separately to either limit to use
            the default behavior.
        flip_canvas : bool, default False
            If True, flip the canvas vertically so that the image looks normal.

        Returns
        -------
        Image
            The image layer.
        """
        layer = _l.Image(
            image, name=name, cmap=cmap, clim=clim, backend=self._get_backend()
        )
        self.add_layer(layer)
        if flip_canvas and not self.y.flipped:
            self.y.flipped = True
        return layer

    def add_layer(
        self,
        layer: _L,
        *,
        over: _l.Layer | Iterable[_l.Layer] | None = None,
        under: _l.Layer | Iterable[_l.Layer] | None = None,
    ) -> _L:
        """Add a layer to the canvas."""
        if over is None and under is None:
            self.layers.append(layer)
        elif over is not None:
            if under is not None:
                raise ValueError("Cannot specify both `over` and `under`")
            if isinstance(over, _l.Layer):
                idx = self.layers.index(over)
            else:
                idx = max([self.layers.index(l) for l in over])
            self.layers.insert(idx + 1, layer)
        else:
            if isinstance(under, _l.Layer):
                idx = self.layers.index(under)
            else:
                idx = min([self.layers.index(l) for l in under])
            self.layers.insert(idx, layer)
        return layer

    @overload
    def group_layers(
        self,
        layers: Iterable[_l.Layer],
        name: str | None = None,
    ) -> _l.LayerGroup: ...

    @overload
    def group_layers(
        self, *layers: _l.Layer, name: str | None = None
    ) -> _l.LayerGroup: ...

    def group_layers(self, layers, *more_layers, name=None):
        """
        Group layers.

        Parameters
        ----------
        layers : iterable of Layer
            Layers to group.

        Returns
        -------
        LayerGroup
            The grouped layer.
        """
        if more_layers:
            if not isinstance(layers, _l.Layer):
                raise TypeError("No overload matches the arguments")
            layers = [layers, *more_layers]
        return _lg.LayerTuple(layers, name=name)

    def _autoscale_for_layer(
        self,
        layer: _l.Layer,
        pad_rel: float | None = None,
        maybe_empty: bool = True,
    ):
        """This function will be called when a layer is inserted to the canvas."""
        if not self.autoscale_enabled:
            return
        if pad_rel is None:
            pad_rel = 0 if layer._NO_PADDING_NEEDED else 0.025
        xmin, xmax, ymin, ymax = layer.bbox_hint()
        if len(self.layers) > 1 or not maybe_empty:
            # NOTE: if there was no layer, so backend may not have xlim/ylim,
            # or they may be set to a default value.
            _xmin, _xmax = self.x.lim
            _ymin, _ymax = self.y.lim
            _dx = (_xmax - _xmin) * pad_rel
            _dy = (_ymax - _ymin) * pad_rel
            xmin = np.min([xmin, _xmin + _dx])
            xmax = np.max([xmax, _xmax - _dx])
            ymin = np.min([ymin, _ymin + _dy])
            ymax = np.max([ymax, _ymax - _dy])

        # this happens when there is <= 1 data
        small_diff = 1e-6
        if np.isnan(xmax) or np.isnan(xmin):
            xmin, xmax = self.x.lim
        elif xmax - xmin < small_diff:
            xmin -= 0.05
            xmax += 0.05
        else:
            dx = (xmax - xmin) * pad_rel
            if (
                xmin != 0
                or not layer._ATTACH_TO_AXIS
                or getattr(layer, "orient", None) is not Orientation.HORIZONTAL
            ):
                xmin -= dx
            xmax += dx
        if np.isnan(ymax) or np.isnan(ymin):
            ymin, ymax = self.y.lim
        elif ymax - ymin < small_diff:
            ymin -= 0.05
            ymax += 0.05
        else:
            dy = (ymax - ymin) * pad_rel
            if (
                ymin != 0
                or not layer._ATTACH_TO_AXIS
                or getattr(layer, "orient", None) is not Orientation.VERTICAL
            ):
                ymin -= dy
            ymax += dy
        self.lims = xmin, xmax, ymin, ymax

    def _cb_overlay_inserted(self, idx: int, layer: _l.Layer):
        _canvas = self._canvas()
        fn = self._get_backend().get("as_overlay")
        for l in _iter_layers(layer):
            _canvas._plt_add_layer(l._backend)
            fn(l._backend, _canvas)
            l._connect_canvas(self)

        if isinstance(layer, _l.LayerWrapper):
            # TODO: check if connecting LayerGroup is necessary
            fn(l._backend, _canvas)
            layer._connect_canvas(self)

    def _cb_removed(self, idx: int, layer: _l.Layer):
        if self._is_grouping:
            return
        _canvas = self._canvas()
        for l in _iter_layers(layer):
            _canvas._plt_remove_layer(l._backend)
            l._disconnect_canvas(self)

    def _cb_layer_grouped(self, group: _l.LayerGroup):
        indices: list[int] = []  # layers to remove
        not_found: list[_l.PrimitiveLayer] = []  # primitive layers to add
        id_exists = set(map(id, self.layers.iter_primitives()))
        for layer in group.iter_children():
            try:
                idx = self.layers.index(layer)
                indices.append(idx)
            except ValueError:
                not_found.extend(_iter_layers(layer))
        if not indices:
            return
        self._is_grouping = True
        try:
            for idx in reversed(indices):
                # remove from the layer list since it is directly grouped
                self.layers.pop(idx)
            self.layers.append(group)
            _canvas = self._canvas()
            for child in not_found:
                if id(child) in id_exists:
                    # skip since it is already in the canvas
                    continue
                child._connect_canvas(self)
                _canvas._plt_add_layer(child._backend)
        finally:
            self._is_grouping = False
        self._cb_reordered()
        self._autoscale_for_layer(group)

    def _generate_colors(self, color: ColorType | None) -> Color:
        if color is None:
            color = self._color_palette.next()
        return color


class Canvas(CanvasBase):
    _CURRENT_INSTANCE: Canvas | None = None

    def __init__(
        self,
        backend: str | None = None,
        *,
        palette: ColormapType | None = None,
    ):
        self._backend = Backend(backend)
        self._backend_object = self._create_backend_object()
        super().__init__(palette=palette)
        self.__class__._CURRENT_INSTANCE = self

    @classmethod
    def from_backend(
        cls,
        obj: protocols.CanvasProtocol,
        *,
        palette: ColormapType | None = None,
        backend: str | None = None,
    ) -> Self:
        """Create a canvas object from a backend object."""
        with patch_dummy_backend() as name:
            # this patch will delay initialization by "_init_canvas" until the backend
            # objects are created.
            self = cls(backend=name, palette=palette)
        self._backend = Backend(backend)
        self._backend_object = obj
        self._init_canvas()
        return self

    def _create_backend_object(self) -> protocols.CanvasProtocol:
        return self._backend.get("Canvas")()

    def _get_backend(self):
        return self._backend

    def _canvas(self) -> protocols.CanvasProtocol:
        return self._backend_object


def _iter_layers(
    layer: _l.Layer,
) -> Iterator[_l.PrimitiveLayer[protocols.BaseProtocol]]:
    if isinstance(layer, _l.PrimitiveLayer):
        yield layer
    elif isinstance(layer, _l.LayerGroup):
        yield from layer.iter_primitive()
    elif isinstance(layer, _l.LayerWrapper):
        yield from _iter_layers(layer._base_layer)
    else:
        raise TypeError(f"Unknown layer type: {type(layer).__name__}")



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
import pyqtgraph as pg
from numpy.typing import NDArray
from qtpy import QtGui

from whitecanvas.backend.pyqtgraph._base import PyQtLayer
from whitecanvas.backend.pyqtgraph._qt_utils import (
    from_qt_line_style,
    to_qt_line_style,
)
from whitecanvas.backend.pyqtgraph.line import MultiLine
from whitecanvas.types import LineStyle


class ArrowHeads(pg.ItemGroup):
    def __init__(self, pos: NDArray[np.floating], angle: NDArray[np.floating]):
        super().__init__()
        for i in range(pos.shape[0]):
            item = pg.ArrowItem(
                pos=pos[i],
                angle=angle[i],
                pen=QtGui.QPen(QtGui.QColor(0, 0, 0, 0)),
                headLen=15,
                pxMode=True,
            )
            self.addItem(item)

    def setData(self, pos: NDArray[np.floating], angle: NDArray[np.floating]):
        ndata = pos.shape[0]
        children = self.childItems()
        scene = self.scene()
        if ndata < len(children):
            for item in self.childItems()[ndata:]:
                scene.removeItem(item)
        elif ndata > len(children):
            for _ in range(ndata - len(children)):
                item = pg.ArrowItem(
                    pen=QtGui.QPen(QtGui.QColor(0, 0, 0, 0)),
                    headLen=15,
                    pxMode=True,
                )
                self.addItem(item)
        for i, item in enumerate(self.childItems()):
            item.setStyle(angle=angle[i])
            item.setPos(*pos[i])

    def setPens(self, pens: list[QtGui.QPen]):
        for pen, item in zip(pens, self.childItems()):
            # item.setPen(pen)
            item.setBrush(QtGui.QBrush(pen.color()))

    if TYPE_CHECKING:

        def childItems(self) -> list[pg.ArrowItem]: ...


class Vectors(pg.ItemGroup, PyQtLayer):
    def __init__(self, x0, dx, y0, dy):
        super().__init__()
        angles = np.degrees(np.arctan2(dy, -dx))
        start = np.column_stack([x0, y0])
        vec = np.column_stack([dx, dy])
        stop = start + vec
        pen = QtGui.QPen(QtGui.QColor(0, 0, 0))
        pen.setCosmetic(True)
        self._brush = QtGui.QBrush(pen.color())
        arrow_heads = ArrowHeads(stop, angles)
        arrow_tails = MultiLine(np.stack([start, stop], axis=1))
        self.addItem(arrow_heads)
        self.addItem(arrow_tails)
        self._start = start
        self._vec = vec
        self._pen_width = pen.widthF()
        self._pen_style = from_qt_line_style(pen.style())
        self._arrow_heads = arrow_heads
        self._arrow_tails = arrow_tails

    def _plt_get_data(self):
        return self._start[:, 0], self._vec[:, 0], self._start[:, 1], self._vec[:, 1]

    def _plt_set_data(self, x0, dx, y0, dy):
        angles = np.degrees(np.arctan2(dy, -dx))
        start = np.column_stack([x0, y0])
        vec = np.column_stack([dx, dy])
        self._arrow_heads.setData(start + vec, angles)
        self._arrow_tails._plt_set_data(np.stack([start, start + vec], axis=1))
        self._start = start
        self._vec = vec

    ##### HasEdges #####
    def _plt_get_edge_width(self) -> float:
        return self._pen_width

    def _plt_set_edge_width(self, width: float):
        self._arrow_tails._plt_set_edge_width(width)
        self._pen_width = width

    def _plt_get_edge_style(self) -> LineStyle:
        return self._pen_style

    def _plt_set_edge_style(self, style: LineStyle):
        self._arrow_tails._plt_set_edge_style(style)
        self._pen_style = to_qt_line_style(style)

    def _plt_get_edge_color(self) -> NDArray[np.float32]:
        return self._arrow_tails._plt_get_edge_color()

    def _plt_set_edge_color(self, color: NDArray[np.float32]):
        self._arrow_tails._plt_set_edge_color(color)
        self._arrow_heads.setPens(self._arrow_tails._qpens)

    def _plt_get_antialias(self) -> bool:
        return True

    def _plt_set_antialias(self, antialias: bool):
        pass



import numpy as np
import pytest

from whitecanvas import new_canvas
from whitecanvas.layers import Layer
from numpy.testing import assert_allclose
from ._utils import assert_color_equal, assert_color_array_equal, filter_warning


def _test_visibility(layer: Layer):
    layer.visible
    layer.visible = False
    layer.visible = True

def test_line(backend: str):
    canvas = new_canvas(backend=backend)
    canvas.add_line(np.arange(10), np.zeros(10))
    layer = canvas.add_line(np.zeros(10))
    canvas.add_line(layer.data)

    repr(layer)
    layer.color
    layer.color = [1.0, 0.0, 0.0, 1.0]
    assert_color_equal(layer.color, "red")
    layer.style
    layer.style = ":"
    assert layer.style == ":"
    layer.width
    layer.width = 2
    assert layer.width == 2
    with pytest.raises(ValueError):
        layer.width = -1
    with pytest.raises(TypeError):
        layer.width = [0, 1]
    with pytest.raises(ValueError):
        layer.data = np.zeros((2, 2, 5))  # 3D data
    with pytest.raises(ValueError):
        layer.data = np.arange(5), np.arange(6)  # shape mismatch
    _test_visibility(layer)
    layer.with_hover_template("x={x:.2f}, y={y:.2f}")
    layer.alpha = 0.5
    layer.alpha
    canvas.add_cdf(np.sqrt(np.arange(20)))
    canvas.add_cdf(np.sqrt(np.arange(20)), orient="horizontal")
    canvas.autoscale()

def test_markers(backend: str):
    canvas = new_canvas(backend=backend)
    canvas.add_markers(np.arange(10), np.zeros(10))
    layer = canvas.add_markers(np.zeros(10))

    repr(layer)
    repr(layer.face)
    repr(layer.edge)
    layer.face.color
    layer.face.alpha
    layer.face.color = [1.0, 0.0, 0.0, 1.0]
    assert_color_equal(layer.face.color, "red")
    layer.face.hatch
    layer.face.hatch = "/"
    assert layer.face.hatch == "/"

    layer.edge.color
    layer.edge.alpha
    layer.edge.color = [0.0, 0.0, 1.0, 1.0]
    assert_color_equal(layer.edge.color, "blue")
    layer.edge.style
    layer.edge.style = ":"
    assert layer.edge.style == ":"
    layer.edge.width
    layer.edge.width = 2
    assert layer.edge.width == 2

    assert layer.symbol == "o"
    layer.symbol = "+"
    assert layer.symbol == "+"

    layer.size
    layer.size = 20
    assert layer.size == 20

    for sym in ["o", "s", "^", "v", "<", ">", "D", "x", "+", "*", "|", "_"]:
        layer.symbol = sym
        assert layer.symbol == sym
    _test_visibility(layer)
    layer.data = np.array([[0, 0], [1, 1], [2, 2]])
    canvas.autoscale()

def test_bars(backend: str):
    canvas = new_canvas(backend=backend)
    canvas.add_bars(np.arange(10), np.zeros(10), bottom=np.ones(10))
    layer = canvas.add_bars(np.arange(10), np.zeros(10))

    repr(layer)
    repr(layer.face)
    repr(layer.edge)
    layer.face.color
    layer.face.alpha
    layer.face.color = [1.0, 0.0, 0.0, 1.0]
    assert_color_equal(layer.face.color, "red")
    layer.face.hatch
    layer.face.hatch = "/"
    assert layer.face.hatch == "/"

    layer.edge.color
    layer.edge.alpha
    layer.edge.color = [0.0, 0.0, 1.0, 1.0]
    assert_color_equal(layer.edge.color, "blue")
    layer.edge.style
    layer.edge.style = ":"
    assert layer.edge.style == ":"
    layer.edge.width
    layer.edge.width = 2
    assert layer.edge.width == 2

    layer.bar_width = 0.5
    assert layer.bar_width == 0.5
    _test_visibility(layer)
    canvas.autoscale()
    assert_allclose(layer.data.x, np.arange(10), rtol=1e-6, atol=1e-6)
    assert_allclose(layer.data.y, np.zeros(10), rtol=1e-6, atol=1e-6)
    layer.data = np.arange(10), np.ones(10)
    layer.bottom = np.arange(10) / 10
    layer.top = np.arange(10) / 10 + 4
    layer.as_edge_only()

def test_step(backend: str):
    canvas = new_canvas(backend=backend)
    layer = canvas.add_step(np.arange(10), np.arange(10), where="pre")

    repr(layer)

    assert_allclose(layer.data.x, np.arange(10))
    assert_allclose(layer.data.y, np.arange(10))
    layer.data = np.arange(10), np.arange(10)
    assert_allclose(layer.data.x, np.arange(10))
    assert_allclose(layer.data.y, np.arange(10))

    layer.where = "mid"
    assert layer.where == "mid"
    assert_allclose(layer.data.x, np.arange(10))
    assert_allclose(layer.data.y, np.arange(10))
    layer.data = np.arange(10), np.arange(10)
    assert_allclose(layer.data.x, np.arange(10))
    assert_allclose(layer.data.y, np.arange(10))

    layer.where = "mid_t"
    assert layer.where == "mid_t"
    assert_allclose(layer.data.x, np.arange(10))
    assert_allclose(layer.data.y, np.arange(10))
    layer.data = np.arange(10), np.arange(10)
    assert_allclose(layer.data.x, np.arange(10))
    assert_allclose(layer.data.y, np.arange(10))

    layer.where = "post"
    assert layer.where == "post"
    assert_allclose(layer.data.x, np.arange(10))
    assert_allclose(layer.data.y, np.arange(10))
    layer.data = np.arange(10), np.arange(10)
    assert_allclose(layer.data.x, np.arange(10))
    assert_allclose(layer.data.y, np.arange(10))

    canvas.add_step(np.arange(10), np.zeros(10), where="mid")
    canvas.add_step(np.arange(10), np.zeros(10), where="post")

def test_infcurve(backend: str):
    canvas = new_canvas(backend=backend)

    layer = canvas.add_infcurve(lambda arr: np.sin(arr / 5))

    layer.color = [1.0, 0.0, 0.0, 1.0]
    assert_color_equal(layer.color, "red")
    layer.style = ":"
    assert layer.style == ":"
    layer.width = 2
    assert layer.width == 2
    _test_visibility(layer)
    layer.with_hover_text("y=sin(x/5)")
    with pytest.raises(TypeError):
        layer.with_hover_text(["x", "y"])
    canvas.layers.remove(layer)  # test disconnection

    layer = canvas.add_infcurve(
        lambda arr, a: np.sin(arr / a)
    ).update_params(a=5)
    canvas.x.lim = (-5, 5)

    # test ufunc
    canvas.add_infcurve(np.sin)

    layer = canvas.add_infline((3, 3), angle=50)
    layer.pos = (2, 2)
    assert layer.pos == (2, 2)
    layer.angle = 45
    assert layer.angle == pytest.approx(45)

    layer.with_hover_text("y=sin(x/5)")
    canvas.x.lim = (-4, 4)
    layer.angle = 90
    assert layer.angle == 90
    canvas.x.lim = (-4, 4)
    canvas.autoscale()
    canvas.layers.remove(layer)  # test disconnection
    canvas.add_hline(1)
    canvas.add_vline(1)

def test_band(backend: str):
    canvas = new_canvas(backend=backend)

    x = np.arange(5)
    layer = canvas.add_band(x, x - 1, x ** 2 / 2)

    repr(layer.face)
    repr(layer.edge)
    layer.face.color
    layer.face.color = [1.0, 0.0, 0.0, 1.0]
    assert_color_equal(layer.face.color, "red")
    layer.face.hatch
    layer.face.hatch = "/"
    assert layer.face.hatch == "/"

    layer.edge.color
    layer.edge.color = [0.0, 0.0, 1.0, 1.0]
    assert_color_equal(layer.edge.color, "blue")
    layer.edge.style
    layer.edge.style = ":"
    assert layer.edge.style == ":"
    layer.edge.width
    layer.edge.width = 2
    assert layer.edge.width == 2
    _test_visibility(layer)
    canvas.autoscale()

def test_image(backend: str):
    canvas = new_canvas(backend=backend)

    rng = np.random.default_rng(0)
    layer = canvas.add_image(rng.random((10, 10)) * 2)

    layer.cmap = "viridis"
    assert layer.cmap == "viridis"
    layer.clim = (0.5, 1.5)
    assert layer.clim == (0.5, 1.5)
    _test_visibility(layer)
    assert layer.data.shape == (10, 10)
    layer.data = np.random.random((10, 10))
    layer.origin = "center"
    layer.shift = (1, 1)
    layer.origin = "edge"
    layer.shift = (-1, -1)
    canvas.autoscale()
    canvas.add_heatmap(rng.random((10, 10)))
    canvas.aspect_ratio = None  # reset

def test_errorbars(backend: str):
    canvas = new_canvas(backend=backend)

    layer = canvas.add_errorbars(np.arange(10), np.zeros(10), np.ones(10))

    layer.color = [1.0, 0.0, 0.0, 1.0]
    assert_color_array_equal(layer.color, "red")
    layer.style = ":"
    assert all(s == ":" for s in layer.style)
    layer.width = 2
    assert all(w == 2 for w in layer.width)
    layer.alpha = 0.5
    layer.alpha
    _test_visibility(layer)

    layer = canvas.add_errorbars(np.arange(10), np.zeros(10), np.ones(10), capsize=0.2)

    layer.color = [1.0, 0.0, 0.0, 1.0]
    assert_color_array_equal(layer.color, "red")
    layer.style = "-."
    assert all(s == "-." for s in layer.style)
    layer.width = 3
    assert all(w == 3 for w in layer.width)
    _test_visibility(layer)
    canvas.autoscale()

def test_texts(backend: str):
    canvas = new_canvas(backend=backend)

    layer = canvas.add_text(np.arange(10), np.zeros(10), list("abcdefghij"))

    repr(layer)
    repr(layer.face)
    repr(layer.edge)
    repr(layer.font)
    layer.anchor
    for anc in [
        "top", "bottom", "left", "right", "center", "top_left", "top_right",
        "bottom_left", "bottom_right"
    ]:
        layer.anchor = anc
        assert layer.anchor == anc

    layer.size
    layer.size = 28
    assert layer.size == 28

    assert layer.ndata == 10
    assert layer.string == list("abcdefghij")
    layer.string = "input-const-text"
    layer.string = list("ABCDEFGHIJ")
    assert layer.string == list("ABCDEFGHIJ")
    layer.face.color
    layer.face.color = "green"
    layer.face.hatch
    layer.face.hatch = "/"
    layer.face.alpha
    layer.face.alpha = 0.6

    layer.edge.color
    layer.edge.color = "blue"
    layer.edge.style
    layer.edge.style = ":"
    layer.edge.width
    layer.edge.width = 2
    layer.edge.alpha
    layer.edge.alpha = 0.6
    _test_visibility(layer)

    layer.pos = np.arange(10) * 2, np.zeros(10)
    assert np.all(layer.pos.x == np.arange(10) * 2)
    assert np.all(layer.pos.y == np.zeros(10))
    layer.rotation = 10
    assert layer.rotation == 10
    layer.color
    layer.color = "red"
    layer.family
    layer.family = "Arial"
    canvas.autoscale()
    canvas.add_text(0, 0, "Hello, World!")
    with filter_warning(backend, ["plotly", "vispy"]):
        layer.with_face(color="red").with_edge(color="blue")
        colors = ["red", "#00FF24"] * 5
        layer.with_face_multi(color=colors).with_edge_multi(width=np.arange(10) / 4)
    layer.data
    layer.data = np.arange(10), np.zeros(10), list("abcdefghij")
    layer.data = np.arange(10) * 2, np.zeros(10), "single"
    layer.set_pos(x=np.arange(10) * 2)

def test_with_text(backend: str):
    canvas = new_canvas(backend=backend)
    x = np.arange(10)
    y = np.sqrt(x)
    canvas.add_line(x, y).with_text([f"{i}" for i in range(10)]).with_text_offset(0.1 ,0.1)
    canvas.add_markers(x, y).with_text([f"{i}" for i in range(10)]).with_text_offset(0.1 ,0.1)
    canvas.add_bars(x, y).with_text([f"{i}" for i in range(10)]).with_text_offset(0.1 ,0.1)
    canvas.add_line(x, y).with_text("x={x:.2f}, y={y:.2f}")
    canvas.add_markers(x, y).with_text("x={x:.2f}, y={y:.2f}")
    canvas.add_bars(x, y).with_text("x={x:.2f}, y={y:.2f}")
    canvas.add_line(x, y).with_markers().with_text([f"{i}" for i in range(10)])
    canvas.add_line(x, y).with_xerr(y/4).with_text([f"{i}" for i in range(10)])
    canvas.add_line(x, y).with_yerr(y/4).with_text([f"{i}" for i in range(10)])
    canvas.add_line(x, y).with_markers().with_text("x={x:2f}, i={i}")
    canvas.add_line(x, y).with_xerr(y/4).with_text("x={x:2f}, i={i}")
    canvas.add_line(x, y).with_yerr(y/4).with_text("x={x:2f}, i={i}")
    canvas.add_markers(x, y).with_network([[0, 1], [4, 3]]).with_text([f"{i}" for i in range(10)])
    canvas.add_markers(x, y).with_network([[0, 1], [4, 3]]).with_text("state-{i}")
    canvas.add_bars(x, y).with_xerr(y/4).with_text([f"{i}" for i in range(10)])
    canvas.add_bars(x, y).with_yerr(y/4).with_text([f"{i}" for i in range(10)])
    canvas.add_bars(x, y).with_xerr(y/4).with_text("{x:1f}, {y:1f},")
    canvas.add_bars(x, y).with_yerr(y/4).with_text("{x:1f}, {y:1f}")
    canvas.autoscale()

def test_rug(backend: str):
    canvas = new_canvas(backend=backend)

    layer = canvas.add_rug(np.arange(10))

    layer.color = [1.0, 0.0, 0.0, 1.0]
    layer.style = ":"
    layer.width = 2
    _test_visibility(layer)
    assert np.all(layer.data == np.arange(10))
    layer.data = np.arange(10, 20)
    assert np.all(layer.data == np.arange(10, 20))
    assert np.all(layer.low == 0)
    assert np.all(layer.high == 1)
    layer.low = 0.5
    layer.high = 1.5
    assert np.allclose(layer.low, 0.5)
    assert np.allclose(layer.high, 1.5)
    layer.update_length(2.0, align="low")
    layer.update_length(np.arange(10) / 5 + 1, align="high")
    layer.update_length(2.0, align="center")
    layer.data = np.random.default_rng(0).normal(size=10)
    with filter_warning(backend, "plotly"):
        layer.color_by_density()
    layer.scale_by_density()
    canvas.autoscale(xpad=(0.01, 0.02), ypad=(0.01, 0.02))


def test_spans(backend: str):
    canvas = new_canvas(backend=backend)

    layer = canvas.add_spans([[5, 10], [18, 30]], orient="vertical")
    canvas.x.lim = (-2, 26)
    layer.data
    layer.data = [[6, 11], [18, 25]]

    layer = canvas.add_spans([[5, 10], [18, 30]], orient="horizontal")
    canvas.y.lim = (-2, 26)
    layer.data
    layer.data = [[6, 11], [18, 25]]

    layer.with_edge("black")

    if backend != "vispy":
        canvas.add_legend()
    canvas.autoscale(xpad=0.01, ypad=0.01)


def test_rects(backend: str):
    canvas = new_canvas(backend=backend)

    layer = canvas.add_rects([[5, 10, 18, 30], [15, 20, 18, 21]])
    layer.data
    assert layer.ndata == 2
    layer.data = [5, 10, 18, 30]
    layer.data = [[5, 10, 18, 30], [15, 20, 18, 21]]
    layer.rects
    layer.as_edge_only()
    layer.with_hover_template("x={left:.2f}, y={bottom:.2f}")


def test_vectors(backend: str):
    if backend not in ("matplotlib", "pyqtgraph", "mock"):
        pytest.skip(f"Vectors not supported by {backend}")
    canvas = new_canvas(backend=backend)
    layer = canvas.add_vectors(np.arange(10), np.zeros(10), np.ones(10), np.zeros(10))
    layer.data
    assert_allclose(layer.data.x, np.arange(10))
    assert_allclose(layer.data.y, np.zeros(10))
    assert_allclose(layer.data.vx, np.ones(10))
    assert_allclose(layer.data.vy, np.zeros(10))
    layer.data = np.arange(10) * 2, np.zeros(10), np.ones(10), np.zeros(10)
    assert_allclose(layer.data.x, np.arange(10) * 2)
    layer.color
    layer.color = [1.0, 0.0, 0.0, 1.0]
    layer.width = 3
    layer.style = ":"



from __future__ import annotations

import numpy as np
from numpy.typing import NDArray

from whitecanvas.layers._primitive import Line, Texts
from whitecanvas.layers.group._collections import LayerContainer
from whitecanvas.types import Alignment


def _norm_bracket_data(
    pos0: tuple[float, float], pos1: tuple[float, float], capsize: float
):
    pos0 = np.array(pos0)
    pos1 = np.array(pos1)
    if np.all(pos0 == pos1):
        raise ValueError("pos0 and pos1 must be different")
    posc = (pos0 + pos1) / 2
    cap_vec = (pos1[::-1] - pos0[::-1]) * np.array([1, -1])
    pos0_cap = pos0 + cap_vec * capsize
    pos1_cap = pos1 + cap_vec * capsize
    line_data = np.stack([pos0_cap, pos0, pos1, pos1_cap], axis=0)
    text_pos = posc - cap_vec * capsize
    return line_data, text_pos


class BracketText(LayerContainer):
    """
    A group of shaped bracket and text.

    This layer group is useful for such as annotating p-values.

       text
    ┌────────┐
    """

    def __init__(
        self,
        pos0: tuple[float, float],
        pos1: tuple[float, float],
        string: str = "",
        capsize: float = 0.1,
        name: str | None = None,
    ):
        line_data, text_pos = _norm_bracket_data(pos0, pos1, capsize)
        text = Texts(*text_pos, string)
        line = Line(
            line_data[:, 0], line_data[:, 1], name="bracket", width=1, color="black"
        )
        super().__init__([text, line], name=name)

    @property
    def text(self) -> Texts:
        return self._children[0]

    @property
    def line(self) -> Line:
        return self._children[1]

    @property
    def capsize(self) -> float:
        """Cap size of the bracket."""
        return float(np.sqrt(np.sum((self.line.data[0] - self.line.data[1]) ** 2)))

    @capsize.setter
    def capsize(self, size: float):
        pos0 = self.line.data[1]
        pos1 = self.line.data[2]
        line_data, text_pos = _norm_bracket_data(pos0, pos1, size)
        self.line.set_data(line_data[:, 0], line_data[:, 1])
        self.text.pos = text_pos


class Panel(LayerContainer):
    """
    A rectangle titled with a text.

        title
    ┌───────────┐
    │           │
    │           │
    └───────────┘
    """

    def __init__(
        self,
        origin: tuple[float, float],
        width: float,
        height: float,
        *,
        title: str = "",
        name: str | None = None,
    ):
        if width <= 0 or height <= 0:
            raise ValueError("width and height must be positive")
        bl = np.array(origin)
        tl = bl + np.array([0, height])
        br = bl + np.array([width, 0])
        tr = bl + np.array([width, height])
        text_pos = tl + np.array([width / 2, 0])
        text = Texts(*text_pos, title, anchor=Alignment.BOTTOM)
        line_data = np.stack([tl, tr, br, bl, tl], axis=0)
        line = Line(
            line_data[:, 0], line_data[:, 1], name="panel", width=1, color="black"
        )
        super().__init__([text, line], name=name)

    @property
    def text(self) -> Texts:
        """Text layer of this panel."""
        return self._children[0]

    @property
    def line(self) -> Line:
        """Line layer of this panel."""
        return self._children[1]

    @property
    def top_left(self) -> NDArray[np.floating]:
        """(x, y) of the top left corner of the panel."""
        return self.line.data.stack()[0]

    @property
    def top_right(self) -> NDArray[np.floating]:
        """(x, y) of the top right corner of the panel."""
        return self.line.data.stack()[1]

    @property
    def bottom_right(self) -> NDArray[np.floating]:
        """(x, y) of the bottom right corner of the panel."""
        return self.line.data.stack()[2]

    @property
    def bottom_left(self) -> NDArray[np.floating]:
        """(x, y) of the bottom left corner of the panel."""
        return self.line.data.stack()[3]

    @property
    def center(self) -> tuple[float, float]:
        """(x, y) of the center of the panel."""
        return (self.top_left + self.bottom_right) / 2

    @center.setter
    def center(self, pos: tuple[float, float]):
        dr = np.array(pos) - self.center
        line_data = self.line.data
        self.line.set_data(line_data.x + dr[0], line_data.y + dr[1])
        self.text.pos = self.text.pos + dr

    @property
    def width(self) -> float:
        """Width of the panel."""
        return self.top_right[0] - self.top_left[0]

    @width.setter
    def width(self, width: float):
        line_data = self.line.data
        if width <= 0:
            raise ValueError("width must be positive")
        w = width / 2
        dx = np.array([-w, w, w, -w, -w])
        self.line.set_data(xdata=line_data.x + dx)

    @property
    def height(self) -> float:
        """Height of the panel."""
        return self.bottom_left[1] - self.top_left[1]

    @height.setter
    def height(self, height: float):
        line_data = self.line.data
        if height <= 0:
            raise ValueError("height must be positive")
        h = height / 2
        dy = np.array([h, h, -h, -h, h])
        self.line.set_data(ydata=line_data.y + dy)
        self.text.pos = self.text.pos + np.array([0, h])



# This example shows how to use the `with_yfill` method to fill the area between
# a line and the x-axis.

import numpy as np

from whitecanvas import new_canvas


def main():
    canvas = new_canvas("matplotlib:qt")
    x = np.linspace(0, 1, 200)
    y = 1 / (1 + x**2)
    canvas.add_line(x, y, color="black", width=2)
    canvas.add_line(x[50:120], y[50:120], color="red", width=4).with_yfill(alpha=0.3)

    canvas.show(block=True)

if __name__ == "__main__":
    main()



from __future__ import annotations

import weakref
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Generic, Iterable, Iterator, TypeVar

import numpy as np
from numpy.typing import NDArray
from psygnal import Signal, SignalGroup

from whitecanvas.backend import Backend
from whitecanvas.layers._legend import EmptyLegendItem, LegendItem
from whitecanvas.protocols import BaseProtocol

if TYPE_CHECKING:
    from typing_extensions import Self

    from whitecanvas.canvas import CanvasBase

_P = TypeVar("_P", bound=BaseProtocol)
_L = TypeVar("_L", bound="Layer")
_T = TypeVar("_T")


class LayerEvents(SignalGroup):
    data = Signal(check_nargs_on_connect=False)  # (data)
    name = Signal(str)
    visible = Signal(bool)
    _layer_grouped = Signal(object)  # (group)


def _no_ref() -> None:
    return None


class Layer(ABC):
    events: LayerEvents
    _events_class: type[LayerEvents]
    _ATTACH_TO_AXIS = False
    _NO_PADDING_NEEDED = False

    def __init__(self, name: str | None = None):
        if not hasattr(self.__class__, "_events_class"):
            self.events = LayerEvents()
        else:
            self.events = self.__class__._events_class()
        self._name = name if name is not None else self.__class__.__name__
        self._x_hint = self._y_hint = None
        self._group_layer_ref: weakref.ReferenceType[LayerGroup] | None = None
        self._canvas_ref = _no_ref

    @property
    @abstractmethod
    def visible(self) -> bool:
        """Return true if the layer is visible"""

    @visible.setter
    def visible(self, visible: bool):
        """Set the visibility of the layer"""

    @property
    def name(self) -> str:
        """Name of this layer."""
        return self._name

    @name.setter
    def name(self, name: str):
        """Set the name of this layer."""
        self._name = str(name)

    def __repr__(self):
        return f"{self.__class__.__name__}<{self.name!r}>"

    def _connect_canvas(self, canvas: CanvasBase):
        """If needed, do something when layer is added to a canvas."""
        self.events._layer_grouped.connect(
            canvas._cb_layer_grouped, unique=True, max_args=1
        )
        self.events.connect(canvas._draw_canvas, unique=True, max_args=0)
        self._canvas_ref = weakref.ref(canvas)

    def _disconnect_canvas(self, canvas: CanvasBase):
        """If needed, do something when layer is removed from a canvas."""
        self.events._layer_grouped.disconnect(canvas._cb_layer_grouped)
        self.events.disconnect(canvas._draw_canvas)
        self._canvas_ref = _no_ref

    def _canvas(self) -> CanvasBase:
        canvas = self._canvas_ref()
        if canvas is None:
            raise ValueError("Layer is not in any canvas.")
        return canvas

    @abstractmethod
    def bbox_hint(self) -> NDArray[np.float64]:
        """Return the bounding box hint (xmin, xmax, ymin, ymax) of this layer."""

    def as_overlay(self) -> Self:
        """Move this layer to the overlay level."""
        canvas = self._canvas()
        canvas.layers.remove(self)
        canvas.overlays.append(self)
        return self

    def _as_legend_item(self) -> LegendItem:
        """Return the legend item for this layer."""
        return EmptyLegendItem()


class PrimitiveLayer(Layer, Generic[_P]):
    """Layers that are composed of a single component."""

    _backend: _P
    _backend_name: str
    _backend_class_name: str | None = None

    @property
    def visible(self) -> bool:
        """Return true if the layer is visible"""
        return self._backend._plt_get_visible()

    @visible.setter
    def visible(self, visible: bool):
        """Set the visibility of the layer"""
        self._backend._plt_set_visible(visible)
        self.events.visible.emit(visible)

    def _create_backend(self, backend: Backend, *args) -> _P:
        """Create a backend object."""
        if self._backend_class_name is not None:
            if "." not in self._backend_class_name:
                backend_cls_name = self._backend_class_name
                self._backend_name = backend.name
                return backend.get(backend_cls_name)(*args)
            else:
                _mod, _cls = self._backend_class_name.rsplit(".")
                backend_cls = getattr(backend.get_submodule(_mod), _cls)
                self._backend_name = backend.name
                return backend_cls(*args)
        for mro in reversed(type(self).__mro__):
            name = mro.__name__
            if (
                issubclass(mro, PrimitiveLayer)
                and mro is not PrimitiveLayer
                and backend.has(name)
            ):
                self._backend_name = backend.name
                return backend.get(name)(*args)
        raise TypeError(
            f"Cannot create a {backend.name} backend for {type(self).__name__}"
        )

    def bbox_hint(self) -> NDArray[np.float64]:
        """Return the bounding box hint (xmin, xmax, ymin, ymax) of this layer."""
        if self._x_hint is None:
            _x = (np.nan, np.nan)
        else:
            _x = self._x_hint
        if self._y_hint is None:
            _y = (np.nan, np.nan)
        else:
            _y = self._y_hint
        return np.array(_x + _y, dtype=np.float64)


class DataBoundLayer(PrimitiveLayer[_P], Generic[_P, _T]):
    @abstractmethod
    def _get_layer_data(self) -> _T:
        """Get the data for this layer."""

    @abstractmethod
    def _set_layer_data(self, data: _T):
        """Set the data for this layer."""

    def _norm_layer_data(self, data: Any) -> _T:
        """Normalize the data for this layer."""
        return data

    @property
    def data(self) -> _T:
        """Data for this layer."""
        return self._get_layer_data()

    @data.setter
    def data(self, data):
        """Set the data for this layer."""
        data_normed = self._norm_layer_data(data)
        self._set_layer_data(data_normed)
        self.events.data.emit(data_normed)


class HoverableDataBoundLayer(DataBoundLayer[_P, _T]):
    def with_hover_text(self, text: str | Iterable[Any]) -> Self:
        """Add hover text to the data points."""
        if isinstance(text, str):
            texts = [text] * self.ndata
        else:
            texts = [str(t) for t in text]
        if len(texts) != self.ndata:
            raise ValueError(
                "Expected text to have the same size as the data, "
                f"got {len(texts)} and {self.ndata}"
            )
        self._backend._plt_set_hover_text(texts)
        return self

    @property
    @abstractmethod
    def ndata(self) -> int:
        """Number of data points."""


class LayerGroup(Layer):
    """A group of layers that will be treated as a single layer in the canvas."""

    def __init__(self, name: str | None = None):
        super().__init__(name)
        self._visible = True

    @abstractmethod
    def iter_children(self) -> Iterator[Layer]:
        """Iterate over all children."""

    def iter_primitive(self) -> Iterator[PrimitiveLayer[BaseProtocol]]:
        for child in self.iter_children():
            if isinstance(child, LayerGroup):
                yield from child.iter_primitive()
            elif isinstance(child, LayerWrapper):
                if isinstance(child._base_layer, LayerGroup):
                    yield from child._base_layer.iter_primitive()
                else:
                    yield child._base_layer
            else:
                yield child

    def _emit_layer_grouped(self):
        """Emit all the grouped signal."""
        for c in self.iter_children():
            c.events._layer_grouped.emit(self)

    @property
    def visible(self) -> bool:
        """Return true if the layer is visible"""
        return self._visible

    @visible.setter
    def visible(self, visible: bool):
        """Set the visibility of the layer"""
        self._visible = visible
        for child in self.iter_children():
            child.visible = visible
        self.events.visible.emit(visible)

    @property
    def _backend_name(self) -> str:
        """The backend name of this layer group."""
        for child in self.iter_children():
            return child._backend_name
        raise RuntimeError(f"No backend name found for {self!r}")

    def bbox_hint(self) -> NDArray[np.float64]:
        """
        Return the bounding box hint (xmin, xmax, ymin, ymax) of this group.

        Note that unless children notifies the change of their bounding box hint, bbox
        hint needs recalculation.
        """
        hints = [child.bbox_hint() for child in self.iter_children()]
        if len(hints) == 0:
            return np.array([np.nan, np.nan, np.nan, np.nan], dtype=np.float64)
        ar = np.stack(hints, axis=1)
        # Any of the four corners could be all-nan. In that case, we should return nan.
        # Otherwise, we should return the known min/max.
        allnan = np.isnan(ar).all(axis=1)
        xmin = np.nan if allnan[0] else np.nanmin(ar[0, :])
        xmax = np.nan if allnan[1] else np.nanmax(ar[1, :])
        ymin = np.nan if allnan[2] else np.nanmin(ar[2, :])
        ymax = np.nan if allnan[3] else np.nanmax(ar[3, :])
        return np.array([xmin, xmax, ymin, ymax], dtype=np.float64)


class LayerWrapper(Layer, Generic[_L]):
    def __init__(
        self,
        base_layer: _L,
    ):
        self._base_layer = base_layer
        super().__init__(base_layer.name)

    @property
    def visible(self) -> bool:
        """Whether the layer is visible."""
        return self._base_layer.visible

    @visible.setter
    def visible(self, visible: bool):
        self._base_layer.visible = visible

    @property
    def name(self) -> str:
        """Name of the layer."""
        return self._base_layer.name

    @name.setter
    def name(self, name: str):
        self._base_layer.name = name

    @property
    def base(self) -> _L:
        """The base layer."""
        return self._base_layer

    def bbox_hint(self) -> NDArray[np.floating]:
        """Return the bounding box hint using the base layer."""
        return self._base_layer.bbox_hint()

    def _connect_canvas(self, canvas: CanvasBase):
        self._base_layer._connect_canvas(canvas)
        return super()._connect_canvas(canvas)

    def _disconnect_canvas(self, canvas: CanvasBase):
        self._base_layer._disconnect_canvas(canvas)
        return super()._disconnect_canvas(canvas)

    @property
    def _ATTACH_TO_AXIS(self) -> bool:
        return self._base_layer._ATTACH_TO_AXIS

    @property
    def _NO_PADDING_NEEDED(self) -> bool:
        return self._base_layer._NO_PADDING_NEEDED

    def _as_legend_item(self) -> LegendItem:
        """Return the legend item for this layer."""
        return self._base_layer._as_legend_item()



from __future__ import annotations

import numpy as np
from cmap import Colormap
from matplotlib.backend_bases import RendererBase
from matplotlib.image import BboxImage
from matplotlib.transforms import Affine2D, Bbox, TransformedBbox

from whitecanvas.backend.matplotlib._base import MplLayer
from whitecanvas.protocols import ImageProtocol, check_protocol


@check_protocol(ImageProtocol)
class Image(BboxImage, MplLayer):
    bbox: TransformedBbox

    def __init__(self, data: np.ndarray):
        if data.ndim == 2:
            shape = data.shape
        else:
            shape = data.shape[:2]
        self._image_shape_2d = shape
        super().__init__(self._make_bbox(), origin="lower")
        self.set_data(data)
        self._cmap = Colormap("gray")
        self._image_transform = Affine2D()

    def _make_bbox(self):
        h, w = self._image_shape_2d
        return Bbox.from_bounds(-0.5, -0.5, w, h)

    def post_add(self, canvas):
        self.bbox = canvas._axes.transData

    def _plt_get_data(self) -> np.ndarray:
        return self.get_array()

    def _plt_set_data(self, data: np.ndarray):
        self.set_data(data)
        if data.ndim == 2:
            shape = data.shape
        else:
            shape = data.shape[:2]
        self._image_shape_2d = shape

    def _plt_get_colormap(self) -> Colormap:
        return self._cmap

    def _plt_set_colormap(self, cmap: Colormap):
        self._cmap = cmap
        self.set_cmap(cmap.to_matplotlib())

    def _plt_get_clim(self) -> tuple[float, float]:
        return self.get_clim()

    def _plt_set_clim(self, clim: tuple[float, float]):
        self.set_clim(clim)

    def _plt_get_scale(self) -> tuple[float, float]:
        mtx = self._image_transform.get_matrix()
        return mtx[0, 0], mtx[1, 1]

    def _plt_set_scale(self, scale: tuple[float, float]):
        mtx = self._image_transform.get_matrix()
        mtx[0, 0], mtx[1, 1] = scale
        self._image_transform = Affine2D(mtx)

    def _plt_get_translation(self) -> tuple[float, float]:
        mtx = self._image_transform.get_matrix()
        return mtx[0, 2], mtx[1, 2]

    def _plt_set_translation(self, translation: tuple[float, float]):
        mtx = self._image_transform.get_matrix()
        mtx[0, 2], mtx[1, 2] = translation
        self._image_transform = Affine2D(mtx)

    def get_window_extent(self, renderer: RendererBase | None = ...) -> Bbox:
        return (
            self._make_bbox().transformed(self._image_transform).transformed(self.bbox)
        )

    def make_image(
        self,
        renderer: RendererBase,
        magnification: float = 1,
        unsampled: bool = False,
    ) -> tuple[np.ndarray, float, float, Affine2D]:
        img, x, y, trans = super().make_image(renderer, magnification, unsampled)
        mtx = self.get_transform().get_matrix()
        if mtx[0, 0] < 0:
            img = img[:, ::-1]
        if mtx[1, 1] < 0:
            img = img[::-1, :]
        return img, x, y, trans



from typing import Callable, TypeVar

from whitecanvas.protocols.canvas_protocol import (
    AxisProtocol,
    CanvasGridProtocol,
    CanvasProtocol,
    HasVisibility,
    TextLabelProtocol,
    TicksProtocol,
)
from whitecanvas.protocols.layer_protocols import (
    BandProtocol,
    BarProtocol,
    BaseProtocol,
    ErrorbarProtocol,
    ImageProtocol,
    LineProtocol,
    MarkersProtocol,
    MeshProtocol,
    MultiLineProtocol,
    RangeDataProtocol,
    TextProtocol,
    VectorsProtocol,
)

__all__ = [
    "VectorsProtocol",
    "BaseProtocol",
    "LineProtocol",
    "MultiLineProtocol",
    "MarkersProtocol",
    "BarProtocol",
    "BandProtocol",
    "ErrorbarProtocol",
    "TextProtocol",
    "ImageProtocol",
    "RangeDataProtocol",
    "CanvasProtocol",
    "CanvasGridProtocol",
    "HasVisibility",
    "TextLabelProtocol",
    "TicksProtocol",
    "AxisProtocol",
    "MeshProtocol",
]

_C = TypeVar("_C")


def check_protocol(p: "type[BaseProtocol]") -> Callable[[_C], _C]:
    def _inner(c):
        if not isinstance(c, p):
            missing = set(filter(lambda x: x.startswith("_plt"), dir(p))) - set(dir(c))
            raise TypeError(
                f"{c!r} does not implement {p.__name__}. \n"
                f"Missing methods: {missing!r}"
            )
        return c

    return _inner



import whitecanvas as wc
from ._utils import assert_color_equal
import pytest


def test_context():
    with wc.theme.context("dark") as theme:
        theme.foreground_color = "yellow"
        assert_color_equal(theme.foreground_color, "yellow")
        canvas = wc.new_canvas("matplotlib:qt")
        assert_color_equal(canvas.x.color, "yellow")
    canvas = wc.new_canvas("matplotlib:qt")
    assert_color_equal(canvas.x.ticks.color, "black")

def test_validator():
    theme = wc.theme.get_theme()

    with pytest.raises(TypeError):
        theme.font.color = 1
    with pytest.raises(TypeError):
        theme.canvas_size = (-1, 2)
    with pytest.raises(Exception):
        theme.line.style = "not-a-style"

def test_update():
    wc.theme.update_default("dark")
    assert_color_equal(wc.theme.get_theme().background_color, "black")

    with wc.theme.update_default("light") as theme:
        theme.background_color = "#EEEEEE"
    assert_color_equal(wc.theme.get_theme().background_color, "#EEEEEE")



from __future__ import annotations

from typing import Any, TypeVar

import numpy as np
from numpy.typing import NDArray

from whitecanvas.backend import Backend
from whitecanvas.layers import _legend
from whitecanvas.layers._primitive.vectors import VectorsLayerEvents
from whitecanvas.layers._sizehint import xyz_size_hint
from whitecanvas.layers.layer3d._base import DataBoundLayer3D
from whitecanvas.protocols import VectorsProtocol
from whitecanvas.types import (
    ArrayLike1D,
    ColorType,
    LineStyle,
    XYZVectorData,
    _Void,
)
from whitecanvas.utils.normalize import as_array_1d, as_color_array
from whitecanvas.utils.type_check import is_real_number

_void = _Void()
_V = TypeVar("_V", bound=VectorsProtocol)


class Vectors3D(DataBoundLayer3D[_V, XYZVectorData]):
    _backend_class_name = "components3d.Vectors3D"
    events: VectorsLayerEvents
    _events_class = VectorsLayerEvents

    def __init__(
        self,
        x: ArrayLike1D,
        y: ArrayLike1D,
        z: ArrayLike1D,
        vx: ArrayLike1D,
        vy: ArrayLike1D,
        vz: ArrayLike1D,
        *,
        name: str | None = None,
        color: ColorType = "blue",
        width: float = 1,
        alpha: float = 1.0,
        style: LineStyle | str = LineStyle.SOLID,
        antialias: bool = True,
        backend: Backend | str | None = None,
    ):
        super().__init__(name=name)
        self._backend = self._create_backend(Backend(backend), x, vx, y, vy, z, vz)
        color = as_color_array(color, x.size)
        self.update(
            color=color, width=width, style=style, alpha=alpha, antialias=antialias
        )
        self._x_hint, self._y_hint, self._z_hint = xyz_size_hint(
            np.concatenate([x, x + vx]),
            np.concatenate([y, y + vy]),
            np.concatenate([z, z + vz]),
        )

    def _get_layer_data(self) -> XYZVectorData:
        x, vx, y, vy, z, vz = self._backend._plt_get_data()
        return XYZVectorData(x, y, z, vx, vy, vz)

    def _norm_layer_data(self, data: Any) -> XYZVectorData:
        if isinstance(data, np.ndarray):
            if data.ndim != 2 or data.shape[1] != 6:
                raise ValueError(f"Expected data to be (N, 4), got {data.shape}")
            xdata, ydata, zdata = data[:, 0], data[:, 1], data[:, 2]
            xvec, yvec, zvec = data[:, 2], data[:, 3], data[:, 4]
        else:
            xdata, ydata, zdata, xvec, yvec, zvec = data
            if xdata is None:
                xdata = self.data.x
            else:
                xdata = as_array_1d(xdata)
            if ydata is None:
                ydata = self.data.y
            else:
                ydata = as_array_1d(ydata)
            if zdata is None:
                zdata = self.data.z
            else:
                zdata = as_array_1d(zdata)
        if xdata.size != ydata.size or xdata.size != zdata.size:
            raise ValueError(
                "Expected xdata, ydata and zdata to have the same size, "
                f"got {xdata.size}, {ydata.size} and {zdata.size}."
            )
        return XYZVectorData(xdata, ydata, zdata, xvec, yvec, zvec)

    def _set_layer_data(self, data: XYZVectorData):
        x0, y0, z0, vx, vy, vz = data
        self._backend._plt_set_data(x0, y0, vx, vy)
        self._x_hint, self._y_hint, self._z_hint = xyz_size_hint(
            np.concatenate([x0, x0 + vx]),
            np.concatenate([y0, y0 + vy]),
            np.concatenate([z0, z0 + vz]),
        )

    def set_data(
        self,
        xdata: ArrayLike1D | None = None,
        ydata: ArrayLike1D | None = None,
        zdata: ArrayLike1D | None = None,
        xvec: ArrayLike1D | None = None,
        yvec: ArrayLike1D | None = None,
        zvec: ArrayLike1D | None = None,
    ):
        self.data = xdata, ydata, zdata, xvec, yvec, zvec

    @property
    def ndata(self) -> int:
        """Number of data points."""
        return self.data.x.size

    @property
    def color(self) -> NDArray[np.floating]:
        """Color of the line."""
        return self._backend._plt_get_edge_color()

    @color.setter
    def color(self, color: ColorType):
        col = as_color_array(color, self.ndata)
        self._backend._plt_set_edge_color(col)
        self.events.color.emit(col)

    @property
    def width(self) -> float:
        """Width of the line."""
        return self._backend._plt_get_edge_width()

    @width.setter
    def width(self, width: float):
        if not is_real_number(width):
            raise TypeError(f"Width must be a number, got {type(width)}")
        if width < 0:
            raise ValueError(f"Width must be non-negative, got {width!r}")
        w = float(width)
        self._backend._plt_set_edge_width(w)
        self.events.width.emit(w)

    @property
    def style(self) -> LineStyle:
        """Style of the line."""
        return LineStyle(self._backend._plt_get_edge_style())

    @style.setter
    def style(self, style: str | LineStyle):
        s = LineStyle(style)
        self._backend._plt_set_edge_style(s)
        self.events.style.emit(s.value)

    @property
    def alpha(self) -> float:
        return float(self.color[:, 3])

    @alpha.setter
    def alpha(self, value: float):
        color = self.color.copy()
        color[:, 3] = value
        self.color = color

    @property
    def antialias(self) -> bool:
        """Whether to use antialiasing."""
        return self._backend._plt_get_antialias()

    @antialias.setter
    def antialias(self, antialias: bool) -> None:
        if not isinstance(antialias, bool):
            raise TypeError(f"Expected antialias to be bool, got {type(antialias)}")
        self._backend._plt_set_antialias(antialias)
        self.events.antialias.emit(antialias)

    def update(
        self,
        *,
        color: ColorType | _Void = _void,
        alpha: float | _Void = _void,
        width: float | _Void = _void,
        style: str | _Void = _void,
        antialias: bool | _Void = _void,
    ):
        if color is not _void:
            self.color = color
        if width is not _void:
            self.width = width
        if style is not _void:
            self.style = style
        if alpha is not _void:
            self.alpha = alpha
        if antialias is not _void:
            self.antialias = antialias
        return self

    def _as_legend_item(self) -> _legend.LineLegendItem:
        return _legend.LineLegendItem(self.color[0], self.width, self.style)



from whitecanvas.layers.layer3d._base import DataBoundLayer3D, Layer3D, PrimitiveLayer3D
from whitecanvas.layers.layer3d.line import Line3D
from whitecanvas.layers.layer3d.markers import Markers3D
from whitecanvas.layers.layer3d.mesh import Mesh3D
from whitecanvas.layers.layer3d.vectors import Vectors3D

__all__ = [
    "Layer3D",
    "PrimitiveLayer3D",
    "DataBoundLayer3D",
    "Line3D",
    "Markers3D",
    "Mesh3D",
    "Vectors3D",
]



from __future__ import annotations

from enum import Enum
from typing import Literal, Union


class _StrEnum(Enum):
    def __eq__(self, other):
        if isinstance(other, str):
            if other in self._value_:
                return self.value == other
            elif (_upper := other.upper()) in self._name_:
                return self.name == _upper
            else:
                return False
        return super().__eq__(other)

    def __hash__(self):
        return hash(self.value)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}.{self.name}<{self.value!r}>"


class LineStyle(_StrEnum):
    SOLID = "-"
    DASH = "--"
    DASH_DOT = "-."
    DOT = ":"


class Hatch(_StrEnum):
    SOLID = ""
    HORIZONTAL = "-"
    VERTICAL = "|"
    CROSS = "+"
    DIAGONAL_BACK = "/"
    DIAGONAL_FORWARD = "\\"
    DIAGONAL_CROSS = "x"
    DOTS = "."


class StepStyle(_StrEnum):
    PRE = "pre"
    POST = "post"
    MID = "mid"
    PRE_T = "pre_t"
    POST_T = "post_t"
    MID_T = "mid_t"


StepStyleStr = Literal["pre", "post", "mid", "pre_t", "post_t", "mid_t"]


class Symbol(_StrEnum):
    CIRCLE = "o"
    SQUARE = "s"
    TRIANGLE_UP = "^"
    TRIANGLE_DOWN = "v"
    TRIANGLE_LEFT = "<"
    TRIANGLE_RIGHT = ">"
    DIAMOND = "D"
    CROSS = "x"
    PLUS = "+"
    STAR = "*"
    VBAR = "|"
    HBAR = "_"

    def has_face(self) -> bool:
        return self not in (
            Symbol.CROSS,
            Symbol.PLUS,
            Symbol.VBAR,
            Symbol.HBAR,
        )


class Modifier(_StrEnum):
    """Enum that represents the keyboard modifiers."""

    SHIFT = "shift"
    CTRL = "ctrl"
    ALT = "alt"
    META = "meta"


class MouseButton(_StrEnum):
    """Enum that represents the mouse buttons."""

    NONE = "none"
    LEFT = "left"
    MIDDLE = "middle"
    RIGHT = "right"
    BACK = "back"
    FORWARD = "forward"


class MouseEventType(_StrEnum):
    """Enum that represents the mouse event type."""

    MOVE = "move"
    PRESS = "press"
    RELEASE = "release"
    DOUBLE_CLICK = "double_click"


class Alignment(_StrEnum):
    TOP = "top"
    BOTTOM = "bottom"
    LEFT = "left"
    RIGHT = "right"
    CENTER = "center"
    TOP_LEFT = "top_left"
    TOP_RIGHT = "top_right"
    BOTTOM_LEFT = "bottom_left"
    BOTTOM_RIGHT = "bottom_right"

    @classmethod
    def merge(cls, vertical, horizontal: Alignment) -> Alignment:
        if vertical not in (Alignment.TOP, Alignment.BOTTOM, Alignment.CENTER):
            raise ValueError(f"{vertical} is not a vertical alignment")
        if horizontal not in (Alignment.LEFT, Alignment.RIGHT, Alignment.CENTER):
            raise ValueError(f"{horizontal} is not a horizontal alignment")
        if vertical is Alignment.TOP:
            if horizontal is Alignment.LEFT:
                return Alignment.TOP_LEFT
            elif horizontal is Alignment.RIGHT:
                return Alignment.TOP_RIGHT
            elif horizontal is Alignment.CENTER:
                return Alignment.TOP
            else:
                raise RuntimeError  # unreachable
        elif vertical is Alignment.BOTTOM:
            if horizontal is Alignment.LEFT:
                return Alignment.BOTTOM_LEFT
            elif horizontal is Alignment.RIGHT:
                return Alignment.BOTTOM_RIGHT
            elif horizontal is Alignment.CENTER:
                return Alignment.BOTTOM
            else:
                raise RuntimeError  # unreachable
        elif vertical is Alignment.CENTER:
            return horizontal
        else:
            raise RuntimeError  # unreachable

    def split(self) -> tuple[Alignment, Alignment]:
        """Split the alignment into vertical and horizontal components."""
        name = self.name
        if "TOP" in name:
            vertical = Alignment.TOP
        elif "BOTTOM" in name:
            vertical = Alignment.BOTTOM
        else:
            vertical = Alignment.CENTER
        if "LEFT" in name:
            horizontal = Alignment.LEFT
        elif "RIGHT" in name:
            horizontal = Alignment.RIGHT
        else:
            horizontal = Alignment.CENTER
        return vertical, horizontal


class Location(_StrEnum):
    TOP_CENTER = "top_center"
    BOTTOM_CENTER = "bottom_center"
    CENTER_LEFT = "center_left"
    CENTER_RIGHT = "center_right"
    CENTER = "center"
    TOP_LEFT = "top_left"
    TOP_RIGHT = "top_right"
    BOTTOM_LEFT = "bottom_left"
    BOTTOM_RIGHT = "bottom_right"
    LEFT_SIDE_TOP = "left_side_top"
    LEFT_SIDE_CENTER = "left_side_center"
    LEFT_SIDE_BOTTOM = "left_side_bottom"
    RIGHT_SIDE_TOP = "right_side_top"
    RIGHT_SIDE_CENTER = "right_side_center"
    RIGHT_SIDE_BOTTOM = "right_side_bottom"
    TOP_SIDE_LEFT = "top_side_left"
    TOP_SIDE_CENTER = "top_side_center"
    TOP_SIDE_RIGHT = "top_side_right"
    BOTTOM_SIDE_LEFT = "bottom_side_left"
    BOTTOM_SIDE_CENTER = "bottom_side_center"
    BOTTOM_SIDE_RIGHT = "bottom_side_right"

    @property
    def is_side(self) -> bool:
        """True if the location is in any of the side locations."""
        return "SIDE" in self.name

    @classmethod
    def parse(cls, value: str) -> Location:
        if value in ("top", "bottom", "left", "right"):
            return cls(value + "_center")
        if "-" in value:
            return cls(value.replace("-", "_"))
        return cls(value)


LocationStr = Literal[
    "top_center", "bottom_center", "center_left", "center_right", "center", "top_left",
    "top_right", "bottom_left", "bottom_right", "left_side_top", "left_side_center",
    "left_side_bottom", "right_side_top", "right_side_center", "right_side_bottom",
    "top_side_left", "top_side_center", "top_side_right", "bottom_side_left",
    "bottom_side_center", "bottom_side_right",
]  # fmt: skip


class Orientation(_StrEnum):
    VERTICAL = "vertical"
    HORIZONTAL = "horizontal"

    @classmethod
    def parse(cls, value):
        if isinstance(value, str):
            if value == "v":
                return cls.VERTICAL
            elif value == "h":
                return cls.HORIZONTAL
        return cls(value)

    def transpose(self) -> Orientation:
        """Transpose the orientation."""
        if self is self.VERTICAL:
            return self.HORIZONTAL
        else:
            return self.VERTICAL

    @property
    def is_vertical(self):
        return self is Orientation.VERTICAL

    @property
    def is_horizontal(self):
        return self is Orientation.HORIZONTAL


OrientationLike = Union[Literal["vertical", "horizontal", "v", "h"], Orientation]


class Origin(_StrEnum):
    """
    Enum that define the center of image.

    1-----+  0: corner
    |     |  1: edge
    |  0  |  2: center (center of image)
    |     |
    +-----+
    """

    CORNER = "corner"
    EDGE = "edge"
    CENTER = "center"


class HistogramShape(_StrEnum):
    step = "step"
    polygon = "polygon"
    bars = "bars"


class HistogramKind(_StrEnum):
    count = "count"
    density = "density"
    probability = "probability"
    frequency = "frequency"
    percent = "percent"



from __future__ import annotations

import bokeh.models as bk_models
import numpy as np
from numpy.typing import NDArray

from whitecanvas.backend.bokeh._base import (
    BokehLayer,
    SupportsMouseEvents,
    from_bokeh_line_style,
    to_bokeh_line_style,
)
from whitecanvas.protocols import LineProtocol, MultiLineProtocol, check_protocol
from whitecanvas.types import LineStyle
from whitecanvas.utils.normalize import arr_color, hex_color


@check_protocol(LineProtocol)
class MonoLine(BokehLayer[bk_models.Line], SupportsMouseEvents):
    def __init__(self, xdata, ydata):
        self._data = bk_models.ColumnDataSource(
            data={"x": xdata, "y": ydata, "hovertexts": np.array([""] * len(xdata))}
        )
        self._model = bk_models.Line(x="x", y="y", line_join="round", line_cap="round")
        self._line_style = LineStyle.SOLID
        self._visible = True
        self._line_color = "#0000FF"

    def _plt_get_visible(self) -> bool:
        return self._visible

    def _plt_set_visible(self, visible: bool):
        if visible:
            self._model.line_color = self._line_color
        else:
            self._model.line_color = "#00000000"
        self._visible = visible

    def _plt_get_data(self):
        return self._data.data["x"], self._data.data["y"]

    def _plt_set_data(self, xdata, ydata):
        self._data.data = {"x": xdata, "y": ydata}

    def _plt_get_edge_width(self) -> float:
        return self._model.line_width

    def _plt_set_edge_width(self, width: float):
        self._model.line_width = width

    def _plt_get_edge_style(self) -> LineStyle:
        return self._line_style

    def _plt_set_edge_style(self, style: LineStyle):
        self._model.line_dash = to_bokeh_line_style(style)
        self._line_style = style

    def _plt_get_edge_color(self) -> NDArray[np.float32]:
        return np.array(arr_color(self._line_color))

    def _plt_set_edge_color(self, color: NDArray[np.float32]):
        if self._visible:
            self._model.line_color = hex_color(color)
        self._line_color = hex_color(color)

    def _plt_get_antialias(self) -> bool:
        return self._model.line_join == "round"

    def _plt_set_antialias(self, antialias: bool):
        self._model.line_join = "round" if antialias else "miter"
        self._model.line_cap = "round" if antialias else "butt"


@check_protocol(MultiLineProtocol)
class MultiLine(BokehLayer[bk_models.MultiLine], SupportsMouseEvents):
    def __init__(self, data: list[NDArray[np.number]]):
        xdata: list[NDArray[np.number]] = []
        ydata: list[NDArray[np.number]] = []
        for seg in data:
            xdata.append(seg[:, 0])
            ydata.append(seg[:, 1])
        self._data = bk_models.ColumnDataSource(
            {
                "x": xdata,
                "y": ydata,
                "edge_color": ["blue"] * len(xdata),
                "width": [1.0] * len(xdata),
                "dash": ["solid"] * len(xdata),
                "hovertexts": [""] * len(xdata),
            }
        )
        self._model = bk_models.MultiLine(
            xs="x",
            ys="y",
            line_color="edge_color",
            line_width="width",
            line_dash="dash",
            line_join="round",
            line_cap="round",
        )
        self._visible = True
        self._line_color = "#0000FF"

    def _plt_get_ndata(self):
        return len(self._data.data["x"])

    def _plt_get_visible(self) -> bool:
        return self._visible

    def _plt_set_visible(self, visible: bool):
        if visible:
            self._model.line_color = "edge_color"
        else:
            self._model.line_color = "#00000000"
        self._visible = visible

    def _plt_get_data(self):
        xs, ys = self._data.data["x"], self._data.data["y"]
        out = []
        for x, y in zip(xs, ys):
            out.append(np.stack([x, y], axis=1))
        return out

    def _plt_set_data(self, data):
        xdata = []
        ydata = []
        for seg in data:
            xdata.append(seg[:, 0])
            ydata.append(seg[:, 1])
        ndata = self._plt_get_ndata()
        edge_color = self._data.data["edge_color"]
        width = self._data.data["width"]
        dash = self._data.data["dash"]
        hovertexts = self._data.data["hovertexts"]
        if len(data) < ndata:
            loss = ndata - len(data)
            edge_color = edge_color[:-loss]
            width = width[:-loss]
            dash = dash[:-loss]
            hovertexts = hovertexts[:-loss]
        elif len(data) > ndata:
            if ndata == 0:
                edge_color = ["blue"] * len(data)
                width = [1.0] * len(data)
                dash = ["solid"] * len(data)
                hovertexts = [""] * len(data)
            else:
                gain = len(data) - ndata
                edge_color = edge_color + edge_color[-1] * gain
                width = width + width[-1] * gain
                dash = dash + dash[-1] * gain
                hovertexts = hovertexts + [""] * gain
        data = {
            "x": xdata,
            "y": ydata,
            "edge_color": edge_color,
            "width": width,
            "dash": dash,
            "hovertexts": hovertexts,
        }
        self._data.data.update(data)

    def _plt_get_edge_width(self) -> NDArray[np.floating]:
        return self._data.data["width"]

    def _plt_set_edge_width(self, width: float):
        if np.isscalar(width):
            width = np.full(self._plt_get_ndata(), width)
        self._data.data["width"] = width

    def _plt_get_edge_style(self) -> list[LineStyle]:
        return [from_bokeh_line_style(d) for d in self._data.data["dash"]]

    def _plt_set_edge_style(self, style: LineStyle | list[LineStyle]):
        if isinstance(style, LineStyle):
            style = [style] * self._plt_get_ndata()
        val = [to_bokeh_line_style(s) for s in style]
        self._data.data["dash"] = val

    def _plt_get_edge_color(self) -> NDArray[np.float32]:
        return np.stack([arr_color(c) for c in self._data.data["edge_color"]], axis=0)

    def _plt_set_edge_color(self, color: NDArray[np.float32]):
        if color.ndim == 1:
            color = [hex_color(color)] * self._plt_get_ndata()
        else:
            color = [hex_color(c) for c in color]
        self._data.data["edge_color"] = color

    def _plt_get_antialias(self) -> bool:
        return self._model.line_join == "round"

    def _plt_set_antialias(self, antialias: bool):
        self._model.line_join = "round" if antialias else "miter"
        self._model.line_cap = "round" if antialias else "butt"
